/* asynsearch.c - asynchronous search routines */

/*
 *				  NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */


#include <stdio.h>
#include <isode/quipu/dap2.h>
#include <isode/quipu/ds_search.h>
#include <isode/quipu/list.h>
#include <isode/quipu/read.h>
#include <isode/quipu/add.h>
#include <isode/quipu/remove.h>
#include <isode/quipu/modify.h>
#include <isode/quipu/modifyrdn.h>
#include "bulkloadglobal.h"
#include "bulkloaderrors.h"
#include "asyn.h"
#include "util.h"

int opno = 0;

extern int dsap_ad;
int noOpsAbandoned;
struct namelist * abanlp = NULLLIST;
struct opsPending * ops = NULLOPL, * endops = NULLOPL;

int
countConns()
{
int count;
struct opsPending * oplp;

  for (oplp = ops , count = 0; oplp != NULLOPL; oplp = oplp->next, count++){};
  return count;
}

struct opsPending *
findOp(opnum)
int opnum;
{
struct opsPending * oplp;

  for (oplp = ops; oplp != NULLOPL; oplp = oplp->next)
  {
    if (oplp->opno == opnum)
      return oplp;
  }
  return NULLOPL;
}

int
addToOpList(opnum, dnstr)
int opnum;
char * dnstr;
{
struct opsPending * oplp;

  oplp = oplist_alloc();
  oplp->opno = opnum;

  oplp->orgname = copy_string(dnstr);

  oplp->lp = NULLLIST;
  oplp->next = NULLOPL;

  if (endops != NULLOPL)
    endops->next = oplp;
  else
    ops = oplp;

  endops = oplp;
}

int
removeFromOpList(opnum)
int opnum;
{
struct opsPending * oplp, * prev;

  if (ops == NULLOPL)
  {
    fprintf(stderr, "Error - trying to remove an operation from an empty operation list\n");
    return;
  }
  for (oplp = ops, prev = NULLOPL; oplp != NULLOPL; oplp = oplp->next)
  {
    if (oplp->opno != opnum)
      prev = oplp;
    else
      break;
  }
  if (oplp == NULLOPL)
  {
     fprintf(stderr, "Error - couldn't find operation %d in list of ops pending\n", opnum);
     return;
  }
  if (oplp == endops)
    endops = prev;
  if (prev != NULLOPL)
    prev->next = oplp->next;
  else
    ops = oplp->next;
  free(oplp->orgname);
  free(oplp);
}

int
de_asynSearch(arg, error, opnumber)
struct ds_search_arg * arg;
struct DSError * error;
int * opnumber;
{
struct DAPindication di;

  di.di_type = 0;
  (void) DapSearch(dsap_ad, ++opno, arg, & di, ROS_ASYNC);
  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      *opnumber = -1;
      return NOTOK;
    case DI_ERROR:
      * opnumber = -1;
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }
  *opnumber = opno;
  return OK;
}

int
de_search(arg, error, result)
struct ds_search_arg * arg;
struct DSError * error;
struct ds_search_result * result;
{

int ret;
struct DAPindication di;
struct DAPresult dapres;

  di.di_type = 0;
  ret = DapSearch(dsap_ad, ++opno, arg, & di, ROS_ASYNC);
  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
#ifdef DEBUG
fprintf( stderr," DapSearch failure %d\n", di.di_un.di_un_error.de_err.dse_type );

fprintf( stderr," DapSearch failure %d\n", di.di_un.di_un_error.de_err.dse_un.dse_un_attribute.DSE_at_plist.DSE_at_what );
#endif
    return DS_ERROR_REMOTE;
    default:
      break;
  }

  ret = de_wait_response(error, & dapres, opno, 0);
  if (ret == OK)
    * result = dapres.dr_res.res_sr;

  return ret;
}

int
de_list(arg, error, result)
struct ds_list_arg * arg;
struct DSError * error;
struct ds_list_result * result;
{

int ret;
struct DAPindication di;
struct DAPresult dapres;

  di.di_type = 0;
  ret = DapList(dsap_ad, ++opno, arg, & di, ROS_ASYNC);
  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }

  ret = de_wait_response(error, & dapres, opno, 0);
  if (ret == OK)
    * result = dapres.dr_res.res_ls;
  return ret;
}


int
de_read(arg, error, result)
struct ds_read_arg * arg;
struct DSError * error;
struct ds_read_result * result;
{

int ret;
struct DAPindication di;
struct DAPresult dapres;

  di.di_type = 0;
  ret = DapRead(dsap_ad, ++opno, arg, & di, ROS_ASYNC);
  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }
  ret = de_wait_response(error, & dapres, opno, 0);
  if (ret == OK)
    * result = dapres.dr_res.res_rd;
  return ret;
}



int
de_addentry( arg, error )
struct ds_addentry_arg * arg;
struct DSError * error;
{

int ret;
struct DAPindication di;
struct DAPresult dapres; /* Should not be needed */

  di.di_type = 0;

  ret = DapAddEntry(dsap_ad, ++opno, arg, & di, ROS_ASYNC);

  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }

  ret = de_wait_response(error, & dapres, opno, 0);

  return ret;
}


int
de_removeentry( arg, error )
struct ds_removeentry_arg * arg;
struct DSError * error;
{

int ret;
struct DAPindication di;
struct DAPresult dapres; /* Should not be needed */

  di.di_type = 0;

  ret = DapRemoveEntry(dsap_ad, ++opno, arg, & di, ROS_ASYNC);

  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }

  ret = de_wait_response(error, & dapres, opno, 0);

  return ret;
}


int
de_modifyentry( arg, error )
struct ds_modifyentry_arg * arg;
struct DSError * error;
{

int ret ;
struct DAPindication di;
struct DAPresult dapres; /* Should not be needed */

  di.di_type = 0;

  ret = DapModifyEntry(dsap_ad, ++opno, arg, & di, ROS_ASYNC);


  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }

  ret = de_wait_response(error, & dapres, opno, 0);

  return ret;
}


int
de_modifyrdn( arg, error )
struct ds_modifyrdn_arg * arg;
struct DSError * error;
{

int ret;
struct DAPindication di;
struct DAPresult dapres;  /* Should not be needed */

  di.di_type = 0;

  ret = DapModifyRDN(dsap_ad, ++opno, arg, & di, ROS_ASYNC);

  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }

  ret = de_wait_response(error, & dapres, opno, 0);

  return ret;
}


int
waitSearchResponse(err, res, opnum)
struct DSError * err;
struct ds_search_result * res;
int * opnum;
{
struct DAPresult dapres;
int ret;

  ret = de_wait_response(err, & dapres, 0, opnum); /* 0 means any op will do */
  if (ret == OK)
  {
    * res = dapres.dr_res.res_sr;
  }
  return ret;
}

int
de_wait_response(err, dapres, opwant, opgot)
struct DSError * err;
struct DAPresult * dapres;
int opwant;
int * opgot;
{
struct DAPindication di;

  while (1)
  {
    (void) DapInitWaitRequest(dsap_ad, 2, &di);
    switch(di.di_type)
    {
      case DI_RESULT:
         if ((opwant != 0) && (di.di_result.dr_id != opwant))
	 {
           /* purpose of this is to throw away any rubbish we don't want */
	   continue;
	 }

         *dapres = di.di_result; /* structure copy */

	 if (opgot)
	   *opgot = di.di_result.dr_id;

         return OK;

      case DI_ERROR:
         if ((opwant != 0) && (di.di_error.de_id != opwant))
	 {
           /* purpose of this is to throw away any rubbish we don't want */
	   continue;
	 }

        *err = di.di_error.de_err; /* structure copy */

	if (opgot)
	  *opgot = di.di_error.de_id;



#ifdef DEBUG
fprintf( stderr," DapWait failure %d\n", di.di_un.di_un_error.de_err.dse_type );

fprintf( stderr," DapWait failure %d\n", di.di_un.di_un_error.de_err.dse_un.dse_un_name.DSE_na_problem );
#endif
        return DS_ERROR_REMOTE;

      case DI_PREJECT:
        if (opgot)
          *opgot = di.di_preject.dp_id;
        return PROTREJECT;

      case DI_ABORT:
/*      doInProgress(); */
        continue;
    }
  }
}
