/* filter.c */

/*
 *				  NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */

#include <isode/quipu/util.h>
#include <isode/quipu/ds_search.h>
#include "util.h"
#include "filter.h"
#include "namelist.h"

#include "bulkloadglobal.h"
#include "bulkloadattrnames.h"
#include "configuration.h"

extern  struct	bulkLoadConfigStruct 	bulkLoadConfig ;

struct ds_search_arg sarg;
struct ds_search_result sresult;
struct DSError serror;

struct s_filter *
andfilter()
{
struct s_filter * fp;

	fp = filter_alloc();
	fp->flt_type = FILTER_AND;
	fp->flt_next = NULLFILTER;
	return fp;
}

struct s_filter *
orfilter()
{
struct s_filter * fp;

	fp = filter_alloc();
	fp->flt_type = FILTER_OR;
	fp->flt_next = NULLFILTER;
	return fp;
}

struct s_filter * 
eqfilter(matchtype, type, value)
int matchtype;
char * type, * value;
{
struct s_filter * fp;
AttributeType at;

	fp = filter_alloc();
	fp->flt_type = FILTER_ITEM;
        fp->FUITEM.fi_type = matchtype;
        if( (at = fp->FUITEM.fi_un.fi_un_ava.ava_type = str2AttrT(type)) != NULLAttrT )
	{
	    fp->FUITEM.fi_un.fi_un_ava.ava_value = str2AttrV(value, at->oa_syntax);
	}
	else fprintf( stderr, "FATAL ERROR : attribute type %s is unknown\n", type ) ;
	
	fp->flt_next = NULLFILTER;
	return fp;
}



/* 
 * Create a substring filter.
 */
 
struct s_filter *
subsfilter(substrtype, type, value)
int substrtype;
char * type, * value;
{
struct s_filter * fp;
AttributeType at;
AttributeValue av ;

	fp = filter_alloc();
        fp->flt_type = FILTER_ITEM;
        fp->FUITEM.fi_type = FILTERITEM_SUBSTRINGS;

	at = fp->FUITEM.UNSUB.fi_sub_type = str2AttrT( type );

	switch (substrtype) {
	
		case LEADSUBSTR:
			fp->FUITEM.UNSUB.fi_sub_initial =
		                avs_comp_new(str2AttrV(value, at->oa_syntax));
		        fp->FUITEM.UNSUB.fi_sub_any = NULLAV;
		        fp->FUITEM.UNSUB.fi_sub_final = NULLAV;

			break;

		case TRAILSUBSTR:
			fp->FUITEM.UNSUB.fi_sub_final =
		                avs_comp_new(str2AttrV(value, at->oa_syntax));

		        fp->FUITEM.UNSUB.fi_sub_any = NULLAV;
		        fp->FUITEM.UNSUB.fi_sub_initial = NULLAV;

			break;

		case ANYSUBSTR:

			fp->FUITEM.UNSUB.fi_sub_final = NULLAV ;

		        fp->FUITEM.UNSUB.fi_sub_any = NULLAV; /* default */

			if( (av = str2AttrV(value, at->oa_syntax)) != NULLAttrV )
		            fp->FUITEM.UNSUB.fi_sub_any = avs_comp_new( av );

		        fp->FUITEM.UNSUB.fi_sub_initial = NULLAV;

			break;
	}
	
	fp->flt_next = NULLFILTER;
	return fp;
}

struct s_filter *
presfilter(type)
char * type;
{
struct s_filter * fp;

        fp = filter_alloc();
        fp->flt_type = FILTER_ITEM;
        fp->FUITEM.fi_type = FILTERITEM_PRESENT;
        fp->FUITEM.UNTYPE = str2AttrT(type);
        fp->flt_next = NULLFILTER;
        return fp;
}


void
makePresentFilter( fpp, attrsToBeFoundp )
struct	s_filter ** fpp ;
struct	namelist *  attrsToBeFoundp ;
{
struct	s_filter * fp;
struct	namelist * attrListp;

	*fpp = orfilter();

        for( attrListp = attrsToBeFoundp ; attrListp != NULLLIST; attrListp = attrListp->next)
	{
	    /* Check if this is the first iteration.. */

	    if( attrListp == attrsToBeFoundp )
	    	fp = (*fpp)->FUFILT = presfilter( attrListp -> name ) ;
	    else
		fp = fp -> flt_next = presfilter( attrListp -> name ) ;
	}
}


/*
 * Since we wish to allow for the possibility of commonNames of the form :-
 *
 *	"TITLE INIT SURNAME","INIT SURNAME", "SURNAME INIT" and "TITLE SURNAME INIT"
 *
 */

void
makeNameSearchFilter( fpp, firstInitcp, surnamecp )
struct	s_filter ** fpp ;
char *	firstInitcp, * surnamecp ;
{
struct	s_filter * fp ;
static	AttributeType 	   cnat = NULLAttrT ;
AttributeValue	   av ;
char 	surNameField[ SMALLBUFLEN ] ;

	*fpp = orfilter();

	if(cnat == NULLAttrT)
	    cnat = str2AttrT( BL_COMMON_NAME );

	if( bulkLoadConfig.commonNameSearchFormat == STARINITSTARSPACESURNAMESTAR ||
	    bulkLoadConfig.commonNameSearchFormat == STARINITSTARSURNAMESTAR        )
	{
	    fp = (*fpp)->FUFILT = subsfilter( ANYSUBSTR, BL_COMMON_NAME, firstInitcp ) ;

	    /* Append on a consecutive surname name component.. */
	    if( bulkLoadConfig.commonNameSearchFormat == STARINITSTARSPACESURNAMESTAR )
	    	sprintf( surNameField, " %s", surnamecp ) ;
	    else
	    	sprintf( surNameField, "%s", surnamecp ) ;

	    if( ( av = str2AttrV( surNameField, cnat->oa_syntax ) ) != NULLAttrV )
	    	fp -> FUITEM.UNSUB.fi_sub_any->avseq_next = avs_comp_new( av ) ;
	}
	else
	{
	    /* Swiss format names to be search for.. */

	    /*  Add in a surname name component.. */
	    sprintf( surNameField, "%s", surnamecp ) ;

	    fp = (*fpp)->FUFILT = subsfilter( ANYSUBSTR, BL_COMMON_NAME, surNameField ) ;

	    if( ( av = str2AttrV( firstInitcp, cnat->oa_syntax ) ) != NULLAttrV )
	   	fp -> FUITEM.UNSUB.fi_sub_any->avseq_next = avs_comp_new( av ) ;
	}
	
	return ;
}
