/* modify.c - attribute modification procedures */


/*
 *				  NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */


#include <isode/quipu/util.h>
#include <isode/quipu/modify.h>
#include <isode/quipu/dua.h>

#include "util.h"
#include "bulkloaderrors.h"
#include "bulkloadattrnames.h"
#include "modify.h"
#include "referral.h"
#include "string.h"

extern IFP merge_acl ;

/*
 * Create an entrymod node from character strings..
 */

struct entrymod *
create_attr_mod_from_cp( em_type, attrTypecp, attrValue )
int    em_type ; /* EM_ADDATTRIBUTE, EM_REMOVERATTRIBUTE,
		    EM_ADDVALUES   , EM_REMOVEVALUES     */
char * attrTypecp, * attrValue ;
{
struct entrymod * emp ;
AttributeType at ;
Attr_Sequence * asp ;


	emp = em_alloc() ;

	emp -> em_type = em_type ;

	asp = &(emp -> em_what) ;

	(*asp) = as_comp_alloc() ;

	at = str2AttrT( attrTypecp ) ;
	(*asp)->attr_type  = at ;

	if( em_type != EM_REMOVEATTRIBUTE )
	    (*asp)->attr_value = avs_comp_new(str2AttrV( attrValue, at->oa_syntax ));
	else
	    (*asp)->attr_value = NULLAV ;	    

	(*asp)->attr_link = NULLATTR ;

	(*asp)->attr_acl  = NULLACL_INFO;

	/* Terminate the linked list */
	emp -> em_next = NULLMOD ;

	return( emp );
}


/*
 * Create an entrymod node from structures..
 */

struct entrymod *
create_attr_mod( em_type, at, av )
int    em_type ; /* EM_ADDATTRIBUTE, EM_REMOVERATTRIBUTE,
		    EM_ADDVALUES   , EM_REMOVEVALUES     */
AttributeType	at ;
AttributeValue	av ;
{
struct entrymod * emp ;
Attr_Sequence * asp ;

	emp = em_alloc() ;

	emp -> em_type = em_type ;

	asp = &(emp -> em_what) ;

	(*asp) = as_comp_alloc() ;

	(*asp)->attr_type  = AttrT_cpy( at ) ;

	if( em_type != EM_REMOVEATTRIBUTE )
	    (*asp)->attr_value = avs_comp_new( AttrV_cpy( av ) );
	else
	    (*asp)->attr_value = NULLAV ;

	(*asp)->attr_link = NULLATTR ;

	(*asp)->attr_acl  = NULLACL_INFO;

	/* Terminate the linked list */
	emp -> em_next = NULLMOD ;

	return( emp );
}


/*
 * Create an entrymod node from structures..
 */

struct entrymod *
create_attr_mod_from_AVS( em_type, at, avs )
int    em_type ; /* EM_ADDATTRIBUTE, EM_REMOVERATTRIBUTE,
		    EM_ADDVALUES   , EM_REMOVEVALUES     */
AttributeType	at  ;
AV_Sequence	avs ;
{
struct entrymod * emp ;
Attr_Sequence * asp ;

	emp = em_alloc() ;

	emp -> em_type = em_type ;

	asp = &(emp -> em_what) ;

	(*asp) = as_comp_alloc() ;

	(*asp)->attr_type = AttrT_cpy( at ) ;

	if( em_type != EM_REMOVEATTRIBUTE )
	    (*asp)->attr_value = avs_cpy( avs ) ;
	else
	    (*asp)->attr_value = NULLAV ;

	(*asp)->attr_link = NULLATTR ;

	(*asp)->attr_acl  = NULLACL_INFO;

	/* Terminate the linked list */
	emp -> em_next = NULLMOD ;

	return( emp );
}


/*
 * Link in the entrymod node at the end of the current linked list.
 */

void
append_attr_mod( entrymodpp, appendednodep )
struct entrymod ** entrymodpp ;    /* Initially the base node */
struct entrymod  * appendednodep ;
{
struct entrymod ** empp = entrymodpp ;


	/* Find the last node in the linked list.. */

	while( (*empp) != NULLMOD )
	{
	    empp = &((*empp) -> em_next) ;
	}


	/* Link in the new entrymod node.. */
	(*empp) = appendednodep ;


	/* Note: appendednodep's node is already NULLMOD terminated */

	return ;
}


void
free_attr_mod_list( entrymodp )
struct entrymod * entrymodp ;
{
struct entrymod * emp = entrymodp ;
struct entrymod * last_emp ;

	/* Return the storage within each node.. */

	while( emp != NULLMOD )
	{
	    as_free( emp -> em_what );

	    last_emp = emp ;

	    emp = emp -> em_next ;

	    (void)free( (char *)last_emp ) ;
	}

	return ;
}


int
modifyAttribute( em_type, entryDN, at, av, errorMsgcpp, errorClassp, errorTypep )
int	em_type ;
DN	entryDN ;
AttributeType	at ;
AttributeValue	av ; /* May be NULLAttrV */
char **	errorMsgcpp ;
int  *	errorClassp, * errorTypep ;
{
int ret_val = NOTOK, loopCount = 0 ;

static struct ds_modifyentry_arg marg ;
static struct DSError            merror ;
static CommonArgs mea_common_args = default_common_args ;
struct entrymod   *mea_changesp = NULLMOD ;

	if (rebind() != OK)
	    return NOTOK;

	marg.mea_common = mea_common_args ; /* struct copy */
	marg.mea_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	marg.mea_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;

	/* Define the entry to be updated.. */

	marg.mea_object = entryDN ;

	marg.mea_changes = NULLMOD ;


	mea_changesp = create_attr_mod( em_type, at, av ) ;

	append_attr_mod( &(marg.mea_changes), mea_changesp ) ;


	/* Request the specifed attribute modifications */
	do { 

	    startUnbindTimer() ;
	    ret_val = de_modifyentry( &marg, &merror ) ;
	    stopUnbindTimer() ;
	    
	} while( loopCount++ < 1 && referralToBeProcessed( ret_val, &merror ) == TRUE ) ;


	free_attr_mod_list( marg.mea_changes );


	if (ret_val != OK)
	{
	    (void)errorClassAndType( &merror, em_type, errorClassp, errorTypep, at, av ) ;

	    *errorMsgcpp = error2pstr( &merror ) ; /* ds_error_free() called by ds_error() */

#ifdef DEBUG
	    fprintf( stderr, "%s\n", *errorMsgcpp ) ;
#endif
    	    return NOTOK;
	}
	else *errorMsgcpp = NULLCP ;


	return( ret_val );
}


int
modifyDirectory( baseModNodep, entryDN, merrorpp )
struct 	entrymod   * baseModNodep ;
DN	entryDN ;
struct  DSError ** merrorpp ;
{
int ret_val = NOTOK, loopCount = 0 ;

static 	struct ds_modifyentry_arg marg ;
static 	struct DSError            merror ;
static 	CommonArgs mea_common_args = default_common_args ;
#ifdef DEBUG
char *	errorMsgcp ;
#endif
	if (rebind() != OK)
	    return NOTOK;

	marg.mea_common = mea_common_args ; /* struct copy */
	marg.mea_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	marg.mea_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;

	/* Define the entry to be updated.. */
	marg.mea_object = entryDN ;

	/* Define the modifications required.. */
	marg.mea_changes = baseModNodep ;


	/* Request the specifed attribute modifications */
	do { 

	    startUnbindTimer() ;
	    ret_val = de_modifyentry( &marg, &merror ) ;
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( ret_val, &merror ) == TRUE ) ;


	if (ret_val != OK)
	{
	    *merrorpp = &merror ;

	    /* NOTE: do not even consider using error2pstr() here as QUIPU's 
	       ds_error() frees the structure ! */

    	    return NOTOK;
	}

	return( ret_val );
}




int
modifyAttributeDN( em_type, entryDN, at, av, errorMsgcpp )
int	em_type ;
DN	entryDN ;
AttributeType	at ;
AttributeValue	av ; /* May be NULLAttrV */
char **	errorMsgcpp ;
{
int ret_val = NOTOK, loopCount = 0 ;
static struct ds_modifyentry_arg marg ;
static struct DSError            merror ;
static CommonArgs mea_common_args = default_common_args ;
struct entrymod   *mea_changesp = NULLMOD ;

	if (rebind() != OK)
	    return NOTOK;

	marg.mea_common = mea_common_args ; /* struct copy */
	marg.mea_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	marg.mea_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;


	/* Define the entry to be updated.. */

	marg.mea_object = entryDN ;


	marg.mea_changes = NULLMOD ;


	mea_changesp = create_attr_mod( em_type, at, av ) ;

	append_attr_mod( &(marg.mea_changes), mea_changesp ) ;



	/* Request the specifed attribute modifications */
	do { 

	    startUnbindTimer() ;
	    ret_val = de_modifyentry( &marg, &merror ) ;
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( ret_val, &merror ) == TRUE ) ;


	free_attr_mod_list( marg.mea_changes );



	if (ret_val != OK)
	{
	    *errorMsgcpp = error2pstr( &merror ) ; /* ds_error_free() called by ds_error() */

#ifdef DEBUG
	    fprintf( stderr, "%s\n", *errorMsgcpp ) ;
#endif
    	    return NOTOK;
	}
	else *errorMsgcpp = NULLCP ;


	return( ret_val );
}




/*
 * Prepare and action the list of modifications for the specified
 * directory entry DN..
 */

int
applyModificationsToDNentry( entryDN, base_changesp, errorMsgcpp )
DN	entryDN ;
struct	entrymod * base_changesp ;
char **	errorMsgcpp ;
{
int ret_val = NOTOK, loopCount = 0 ;

static struct ds_modifyentry_arg marg ;
static struct DSError            merror ;
static CommonArgs mea_common_args = default_common_args ;
		
	if (rebind() != OK)
	    return NOTOK;


	marg.mea_common = mea_common_args ; /* struct copy */
	marg.mea_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	marg.mea_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;


	/* Define the entry to be updated.. */
	
	marg.mea_object = entryDN ;


	marg.mea_changes = base_changesp ;
	

	/* Request the specifed attribute modifications */
	do { 

	    startUnbindTimer() ;
	    ret_val = de_modifyentry( &marg, &merror ) ;
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( ret_val, &merror ) == TRUE ) ;


	if (ret_val != OK)
	{
	    *errorMsgcpp = error2pstr( &merror ) ; /* ds_error_free() called by ds_error() */

#ifdef DEBUG
	    fprintf( stderr, "%s\n", *errorMsgcpp ) ;
#endif
    	    return NOTOK;
	}
	else *errorMsgcpp = NULLCP ;

	return( ret_val );
}




/*
 * Prepare and action an attribute value modification request.
 */

int
AddReplaceDNattr( EntryDNstr, attrTypecp, newAttrValue, oldAttrValue, AddOrReplace, errorMsgcpp )
char *	EntryDNstr, *attrTypecp, *newAttrValue, *oldAttrValue ;
int    	AddOrReplace ; /* Either MOD_ADD_DN_ATTR, MOD_ADD_DN_value,
			  MOD_REPLACE_DN_VALUE or MOD_REMOVE_DN_VALUE */
char ** errorMsgcpp ;
{
int ret_val = NOTOK, loopCount = 0 ;

static struct ds_modifyentry_arg marg ;
static struct DSError            merror ;
static CommonArgs mea_common_args = default_common_args ;
struct entrymod   *mea_changesp = NULLMOD ;

		
	if (rebind() != OK)
	    return NOTOK;


	marg.mea_common = mea_common_args ; /* struct copy */
	marg.mea_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	marg.mea_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;


	/* Define the entry to be updated.. */
	
	marg.mea_object = str2dn(EntryDNstr);


	marg.mea_changes = NULLMOD ;

	
	/* Do we need to remove an attribute value ? */

	if( AddOrReplace == MOD_REPLACE_DN_VALUE ||
	    AddOrReplace == MOD_REMOVE_DN_VALUE    )
	{
	    mea_changesp = create_attr_mod_from_cp( EM_REMOVEVALUES, attrTypecp,
	    						oldAttrValue ) ;

	    append_attr_mod( &(marg.mea_changes), mea_changesp ) ;
	}

	/* Do we need to add an attribute value ? */

	if( AddOrReplace == MOD_ADD_DN_VALUE     ||
	    AddOrReplace == MOD_REPLACE_DN_VALUE   )
	{
	    mea_changesp = create_attr_mod_from_cp( EM_ADDVALUES, attrTypecp,
	    						newAttrValue ) ;

	    append_attr_mod( &(marg.mea_changes), mea_changesp ) ;
	}
	

	/* Do we need to add an atribute ? */

	if( AddOrReplace == MOD_ADD_DN_ATTR )
	{
	    mea_changesp = create_attr_mod_from_cp( EM_ADDATTRIBUTE, attrTypecp,
	    						newAttrValue ) ;

	    append_attr_mod( &(marg.mea_changes), mea_changesp ) ;
	}

	

	/* Request the specifed attribute modifications */
	do { 

	    startUnbindTimer() ;
	    ret_val = de_modifyentry( &marg, &merror ) ;
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( ret_val, &merror ) == TRUE ) ;


	free_attr_mod_list( marg.mea_changes );
	dn_free( marg.mea_object ) ;



	if (ret_val != OK)
	{
	    *errorMsgcpp = error2pstr( &merror ) ; /* ds_error_free() called by ds_error() */

#ifdef DEBUG
	    fprintf( stderr, "%s\n", *errorMsgcpp ) ;
#endif    	    
	    return NOTOK;
	}
	else *errorMsgcpp = NULLCP ;

	return( ret_val );
}


/*
 * Given the indication of whether an attribute or an attribute value
 * addition is required select the operation. If the one selected should
 * fail with an error that suggests that the alternative might succeed
 * then attempt it.
 */

void	addAttrOrValue( entryDN, at, av, valueAlreadyPresentForAttrT )
DN		entryDN ;
AttributeType	at ;
AttributeValue	av ;
int		valueAlreadyPresentForAttrT ;
{
int	tryAddAttribute = TRUE, errorClass, errorType ;
char *	errorMsgcp ;

	if( valueAlreadyPresentForAttrT == TRUE )
	{
	    /* Add value */

	    if( modifyAttribute( EM_ADDVALUES, entryDN, at, av, 
		    		 &errorMsgcp, &errorClass, &errorType ) == NOTOK )
	    {
		/* May need to add the attribute instead.. */
		if( errorClass != DSE_ATTRIBUTEERROR || errorType != DSE_AT_NOSUCHATTRIBUTE )
		    tryAddAttribute = FALSE ;
	    	
		free( errorMsgcp ) ;
	    }
	    else tryAddAttribute = FALSE ;
	}

	if( tryAddAttribute == TRUE )
	{
	    /* Add attribute */
	    if( modifyAttribute( EM_ADDATTRIBUTE, entryDN, at, av, 
		    		 &errorMsgcp, &errorClass, &errorType ) == NOTOK )
		free( errorMsgcp ) ;
	}

	return ;
}



/*
 * Depending whether an attribute or an attribute value
 * removal is required select the operation.
 */

void	removeAttrOrValue( entryDN, at, av, removingLastValue )
DN		entryDN ;
AttributeType	at ;
AttributeValue	av ;
int		removingLastValue ;
{
int	errorClass, errorType ;
char *	errorMsgcp ;

	if( removingLastValue == TRUE )
	{
	    /* remove attribute */
	    if( modifyAttribute( EM_REMOVEATTRIBUTE, entryDN, at, av, 
		    		 &errorMsgcp, &errorClass, &errorType ) == NOTOK )
		free( errorMsgcp ) ;
	}
	else
	{
	    /* remove value */
	    if( modifyAttribute( EM_REMOVEVALUES, entryDN, at, av, 
		    		 &errorMsgcp, &errorClass, &errorType ) == NOTOK )
		free( errorMsgcp ) ;
	}

	return ;
}

/*
 * Check if the specified modification on the given Attribute Type
 * is already stored in the modifications linked list.
 * The is intended for ADD ATTRIBUTE and REMOVE ATTRIBUTE modifications only,
 * unless the value is not relevant for an ADD VALUES operation.
 */

int
modificationIsAlreadyListed( baseModNodep, em_type, at )
struct	entrymod   * baseModNodep  ;
int 	em_type ;
AttributeType	at ;
{
int	actionAlreadyInList = FALSE ;
struct	entrymod   * currModNodep = baseModNodep ;

	while( currModNodep != NULLMOD )
	{
	    if( currModNodep -> em_type == em_type  &&
	        currModNodep -> em_what != NULLATTR &&
	        AttrT_cmp( currModNodep -> em_what -> attr_type, at ) == 0 )
	    {
	    	actionAlreadyInList = TRUE ;
		break ;
	    }

	    currModNodep = currModNodep -> em_next ;
	}

	return( actionAlreadyInList ) ;
}
