#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
	  
#include <isode/quipu/name.h>
#include <isode/quipu/util.h>

#include "bulkloadglobal.h"
#include "bulkloaderrors.h"
#include "bulkloadattrnames.h"
#include "configuration.h"
#include "processandupdate.h"
#include "util.h"

#ifdef DEBUG
#include "util.h"
#endif

extern  struct	bulkLoadConfigStruct 	bulkLoadConfig ;

char *	processFileTokens[ NUMPROCESSFILETOKENS ] = {

	    "DRAFTFILE",
	    "DONEFILE",
	    "NOACTIONFILE",
	    "PROBLEMSFILE",
	    "PROCESS",
	    "INVALIDPROCESSING"
	} ;

#define DONEFILETAIL		".done"
#define NOACTIONFILETAIL	".noAction"
#define PROBLEMFILETAIL		".PROBLEM"
#define MAXTAILCOUNT		999


#define MAXLINELEN		2048
#define FILENAMEBUFFLEN		128

#define INITLINEBUFLEN		20000	/* Default attribute line buffer size */
#define LINEBUFSIZEINC		10000   /* By how many octets the line buffer shall
					   be increased each time the buffer fills  
					   NOTE: MUST BE GREATER THAN "MAXLINELEN" 
					   since the realloc will then read in
					   the next line */



#define NOTOKALLDATAREAD	2 /* Must ensure this does not clash
				     with NOTOK or OK. 
				     Indicates that a request to retrieve a rootedAt, 
				     RDN of attribute line failed to find one but no 
				     fatal error was detected which would have been 
				     indicated by a NOTOK */

struct lineBufferStruct {

	char *  lineBufferp ;

	int 	bufferSize  ; /* In characters */
} ;


struct	inFileDetails {

	FILE *	inFilep ; 				/* permanently opened for reading */
	char	inFilename[ SMALLBUFLEN ] ;
	char *  savedLineBuffercp ;			/* During a switch to a parse file the last read
							   but unprocessed inFile line is buffered */

	char	parse1Filename[ SMALLBUFLEN ] ;
	char	parse2Filename[ SMALLBUFLEN ] ;	

	FILE *	parseInputFilep  ;			/* either the inFile or parse1File */
	char	parseInputFilename[ SMALLBUFLEN ] ;
	FILE *	parseOutputFilep ;			/* either parse1File or parse2File */
	char	parseOutputFilename[ SMALLBUFLEN ] ;

#define	SWAPPARSEFILES			0 /* This macro must not conflict with the following three macros.. */

#define	PARSEINPUTFILEINPARSE1OUT	1
#define	PARSE1INPARSE2OUT		2
#define	PARSE2INPARSE1OUT		3
	int	fileUsage ;		/* ...any of the preceding THREE macros */
	
	int	errorOccurred ; 	/* TRUE or FALSE */
	int	fatalErrorOccurred ; 	/* TRUE or FALSE */
} ;


void
closeParseFiles( IFDp )
struct	inFileDetails * IFDp ;
{
	if( IFDp -> parseInputFilep != ((FILE *)NULL) && IFDp -> parseInputFilep != IFDp -> inFilep ) 
	    fclose( IFDp -> parseInputFilep ) ;

	if( IFDp -> parseOutputFilep != ((FILE *)NULL) ) 
	    fclose( IFDp -> parseOutputFilep ) ;

	return ;
}

void
finishIFD( IFDp )
struct	inFileDetails * IFDp ;
{
	closeParseFiles( IFDp ) ;

	if( IFDp -> parse1Filename[ 0 ] != '\0' && strncmp( IFDp -> parse1Filename, PARSEAFILENAME, 12 ) == 0 )
	    unlink( IFDp -> parse1Filename ) ;
	    
	if( IFDp -> parse2Filename[ 0 ] != '\0'  && strncmp( IFDp -> parse2Filename, PARSEAFILENAME, 12 ) == 0 )
	    unlink( IFDp -> parse2Filename ) ;

	if( IFDp -> savedLineBuffercp != NULLCP )
	{
	    free( IFDp -> savedLineBuffercp ) ;
	    IFDp -> savedLineBuffercp = NULLCP ;
	}

	return ;
}

/*
 * Depending on the value of "fileUsage" open to two parse files
 * for reading or writing.
 */

int
openParseFiles( IFDp, fileUsage )
struct	inFileDetails * IFDp ;
int	fileUsage ;	/* PARSEINPUTFILEINPARSE1OUT, PARSE1INPARSE2OUT or PARSE2INPARSE1OUT */
{
int	ret_val = OK ;
	
	closeParseFiles( IFDp ) ;

	
	IFDp -> fileUsage = fileUsage ;

	if( IFDp -> fileUsage == PARSEINPUTFILEINPARSE1OUT ||
	    IFDp -> fileUsage == PARSE2INPARSE1OUT 	     )
	{
	    /* Open the first parse file for writing */
	    if( (IFDp -> parseOutputFilep = fopen( IFDp->parse1Filename, "w" )) == NULL )
	    {
	    	perror( "Attempt to open a bulk load parse file for writing failed" );
	    	ret_val = NOTOK ;
	    }
	    else
	    {
		strcpy( IFDp -> parseOutputFilename, IFDp->parse1Filename ) ;

	    	if( IFDp -> fileUsage == PARSE2INPARSE1OUT )
		{
	    	    if( (IFDp -> parseInputFilep = fopen( IFDp->parse2Filename, "r+" )) == NULL )
		    {
		    	perror( "Attempt to open a bulk load parse file for reading failed" );
			ret_val = NOTOK ;
		    }
		    else strcpy( IFDp -> parseInputFilename, IFDp->parse2Filename ) ;
		}
		else
		{
		    /* fileUsage == PARSEINPUTFILEINPARSE1OUT */
		    IFDp -> parseInputFilep = IFDp -> inFilep ; /* Always open solely for reading */
		    strcpy( IFDp -> parseInputFilename, IFDp -> inFilename ) ;

		    /* parseInputFilename is not required */
		}
	    }
	}
	else
	{
	    /* Open the first parse file for reading */
	    if( (IFDp -> parseInputFilep = fopen( IFDp->parse1Filename, "r+" )) == NULL )
	    {
	    	perror( "Attempt to open a bulk load parse file for reading failed" );
	    	ret_val = NOTOK ;
	    }
	    else 
	    {
		strcpy( IFDp -> parseInputFilename, IFDp->parse1Filename ) ;

	    	if( (IFDp -> parseOutputFilep = fopen( IFDp->parse2Filename, "w" )) == NULL )
		{
	    	    perror( "Attempt to open a bulk load parse file for writing failed" );
		    ret_val = NOTOK ;
		}
		else strcpy( IFDp -> parseOutputFilename, IFDp->parse2Filename ) ;
	    }
	}	

	return( ret_val ) ;
}

void
setUpParseFileDetails( IFDp, lineBufferStructp, fileUsageSpecification )
struct	inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
int	fileUsageSpecification ;
{
int	fileUsage ;

	if( fileUsageSpecification == SWAPPARSEFILES )
	{
	    /* save the currently buffered inFile line (may be empty string) */
	    if( IFDp -> fileUsage == PARSEINPUTFILEINPARSE1OUT )
	    {
	    	if( IFDp -> savedLineBuffercp != NULLCP )
		{
		    free( IFDp -> savedLineBuffercp ) ;
		    IFDp -> savedLineBuffercp = NULLCP ;
		}

		if( lineBufferStructp -> lineBufferp != NULLCP )
		{
		    IFDp -> savedLineBuffercp = strdup( lineBufferStructp -> lineBufferp ) ;
		    *(lineBufferStructp -> lineBufferp) = '\0' ;
		}
	    }
	    else *(lineBufferStructp -> lineBufferp) = '\0' ; /* Will be editing a different file */
	    
	    if( IFDp -> fileUsage == PARSEINPUTFILEINPARSE1OUT ||
	    	IFDp -> fileUsage == PARSE2INPARSE1OUT   )
	    	fileUsage = PARSE1INPARSE2OUT ;
	    else
	    	fileUsage = PARSE2INPARSE1OUT ;

	    (void)openParseFiles( IFDp, fileUsage ) ;
	}
	else if( fileUsageSpecification == PARSEINPUTFILEINPARSE1OUT )
	{
	    fileUsage = PARSEINPUTFILEINPARSE1OUT ;

	    /* Return the saved line now that we are returning to read from the inFile */
	    if( IFDp -> savedLineBuffercp != NULLCP )
	    {
		strcpy( lineBufferStructp -> lineBufferp, IFDp -> savedLineBuffercp ) ;
		free( IFDp -> savedLineBuffercp ) ;
		IFDp -> savedLineBuffercp = NULLCP ;
	    }
	    
	    (void)openParseFiles( IFDp, fileUsage ) ;
	}

#ifdef DEBUG
fprintf( stderr, "LOCATION 1 %ld\n", ftell( IFDp ->inFilep )) ;
fprintf( stderr, "LOCATION 2 %ld\n", ftell( IFDp ->parseInputFilep )) ;
#endif

	
	IFDp -> errorOccurred      = FALSE ;
	IFDp -> fatalErrorOccurred = FALSE ;

	return ;
}

int
initialiseInFiles( IFDp, inFilenamecp, parse1Filenamecp, parse2Filenamecp )
struct	inFileDetails * IFDp ;
char *	inFilenamecp ;
char *	parse1Filenamecp, * parse2Filenamecp ;
{
int	ret_val = NOTOK ;
char *	parse1filecp, parse1fileBuf[ SMALLBUFLEN ], * parse2filecp, parse2fileBuf[ SMALLBUFLEN ] ;

	IFDp -> parseInputFilep  = (FILE *)NULL ;
	IFDp -> parseOutputFilep = (FILE *)NULL ;

	strcpy( IFDp -> inFilename, inFilenamecp ) ;

	IFDp -> savedLineBuffercp = NULLCP ;
	
	IFDp -> fileUsage = -1 ;
	
	IFDp -> parse1Filename[ 0 ] = '\0' ;
	IFDp -> parse2Filename[ 0 ] = '\0' ;
	
	/* Attempt to open the input file for reading.. */
	if( (IFDp -> inFilep = fopen( inFilenamecp, "r" ) ) == NULL )
	{
	    perror( "Attempt to open user input file for reading failed" );
	}
	else
 	{
	    sprintf( parse1fileBuf, "%sXXXXXX", parse1Filenamecp ) ;
	    parse1filecp = mktemp( parse1fileBuf ) ;
	    
	    sprintf( parse2fileBuf, "%sXXXXXX", parse2Filenamecp ) ;
	    parse2filecp = mktemp( parse2fileBuf ) ;
	    
	    strcpy( IFDp -> parse1Filename, parse1filecp ) ;
	    strcpy( IFDp -> parse2Filename, parse2filecp ) ;
	    
	    if( openParseFiles( IFDp, PARSEINPUTFILEINPARSE1OUT ) == OK )
	    	ret_val = OK ;
	}

	IFDp -> errorOccurred = FALSE ;
	IFDp -> fatalErrorOccurred = FALSE ;

	return( ret_val ) ;
}

void
reportParseError( IFDp, errorcp1, errorcp2 )
struct	inFileDetails * IFDp ;
char *	errorcp1, * errorcp2 ;
{
	fprintf( IFDp->parseOutputFilep, "#ERROR ***  %s\n", errorcp1 ) ;

#ifdef DEBUG
fprintf( stderr, "#ERROR %s\n", errorcp1 ) ;
#endif

	if( errorcp2 != NULLCP )
	    fprintf( IFDp -> parseOutputFilep, "#ERROR *** %s\n", errorcp2 ) ;

#ifdef DEBUG
if( errorcp2 != NULLCP )
	fprintf( stderr, "#ERROR %s\n", errorcp2 ) ;
#endif

	IFDp -> errorOccurred = TRUE ;

	return ;
}


void
reportFatalParseError( IFDp, errorcp1, errorcp2 )
struct	inFileDetails * IFDp ;
char *	errorcp1, * errorcp2 ;
{
	fprintf( IFDp->parseOutputFilep, "#ERROR FATAL %s\n", errorcp1 ) ;

	if( errorcp2 != NULLCP )
	    fprintf( IFDp->parseOutputFilep, "#ERROR FATAL%s\n", errorcp2 ) ;

	IFDp -> fatalErrorOccurred = TRUE ;

	return ;
}

/**************************************************************************************/

void
flushOutFiles( OFD )
struct	outputFileDetails OFD[] ;
{
	if( OFD[ OUTPUTDONEFILE ].outFilep != (FILE *)NULL )
	    fflush( OFD[ OUTPUTDONEFILE ].outFilep ) ;

	if( OFD[ OUTPUTNOACTIONFILE ].outFilep != (FILE *)NULL )
	    fflush( OFD[ OUTPUTNOACTIONFILE ].outFilep ) ;

	if( OFD[ OUTPUTPROBLEMFILE ].outFilep != (FILE *)NULL )
	    fflush( OFD[ OUTPUTPROBLEMFILE ].outFilep ) ;

	return ;
}


void
resetOutFileDetails( OFD )
struct	outputFileDetails OFD[] ;
{
int 	f ;

	for( f = MINOUTPUTFILEINDEX ; f <= MAXOUTPUTFILEINDEX ; f++ )
	{
	    OFD[ f ].RootedAtOutput     = FALSE  ;
	    OFD[ f ].RootedAtDNcp	= NULLCP ;
	    OFD[ f ].SubDirListResolvedParentDNcp = NULLCP ;
	    
	    OFD[ f ].RDNoutput     	= FALSE  ;
	    OFD[ f ].rdnLinecp		= NULLCP ;
	}

	return ;
}

void
prepareOutFiles( OFD, inFileType, sourceNamecp )
struct	outputFileDetails OFD[] ;
int	inFileType ;
char *	sourceNamecp ;
{
int 	f ;

	for( f = MINOUTPUTFILEINDEX ; f <= MAXOUTPUTFILEINDEX ; f++ )
	{
	    if( f != OUTPUTDRAFTFILE ) /* Do not do anything to the draft file ! */
	    {
	    	if( inFileType == NEWDMFORMATINPUT )
	    	{
	    	    fprintf( OFD[ f ].outFilep, "%s\n\n", NEWDMFORMATFIRSTLINE ) ;
		}

		fprintf( OFD[ f ].outFilep, "%s%s\n\n", SOURCELINEHEAD, sourceNamecp ) ;
	    }
	}
	
	return ;
}

void
setUpOutFilesRA( OFD, rootedAtDNcp )
struct	outputFileDetails OFD[] ;
char *	rootedAtDNcp ;
{
int 	f ;

	for( f = MINOUTPUTFILEINDEX ; f <= MAXOUTPUTFILEINDEX ; f++ )
	{
	    if( OFD[ f ].RootedAtDNcp != NULLCP )
	    	free( OFD[ f ].RootedAtDNcp ) ;

	    if( rootedAtDNcp != NULLCP )
	        OFD[ f ].RootedAtDNcp = strdup( rootedAtDNcp ) ;
	    else
		OFD[ f ].RootedAtDNcp = NULLCP ;

	    if( OFD[ f ].SubDirListResolvedParentDNcp != NULLCP )
	    {
	    	free( OFD[ f ].SubDirListResolvedParentDNcp ) ;
	    	OFD[ f ].SubDirListResolvedParentDNcp = NULLCP ;
	    }

	    OFD[ f ].RootedAtOutput = FALSE ;
	}
	
	return ;
}

void
setUpOutFilesSubDirListResolvedDN( OFD, SubDirListResolvedParentDNcp )
struct	outputFileDetails OFD[] ;
char *	SubDirListResolvedParentDNcp ;
{
int 	f ;

	for( f = MINOUTPUTFILEINDEX ; f <= MAXOUTPUTFILEINDEX ; f++ )
	{
	    if( SubDirListResolvedParentDNcp != NULLCP && OFD[ f ].SubDirListResolvedParentDNcp == NULLCP ||
	        SubDirListResolvedParentDNcp == NULLCP && OFD[ f ].SubDirListResolvedParentDNcp != NULLCP ||
		(SubDirListResolvedParentDNcp != NULLCP && OFD[ f ].SubDirListResolvedParentDNcp != NULLCP &&
			strcmp( SubDirListResolvedParentDNcp, OFD[ f ].SubDirListResolvedParentDNcp ) != 0 ) )
		OFD[ f ].RootedAtOutput = FALSE ; /* Change in resolved parent DN */

	    if( OFD[ f ].SubDirListResolvedParentDNcp != NULLCP )
	    	free( OFD[ f ].SubDirListResolvedParentDNcp ) ;

	    if( SubDirListResolvedParentDNcp != NULLCP )
	    	OFD[ f ].SubDirListResolvedParentDNcp = strdup( SubDirListResolvedParentDNcp ) ;
	    else
	    	OFD[ f ].SubDirListResolvedParentDNcp = NULLCP ;
	}

	return ;
}

void
setUpOutFilesRDN( OFD, rdnLinecp )
struct	outputFileDetails OFD[] ;
char *	rdnLinecp ;
{
int 	f ;

	for( f = MINOUTPUTFILEINDEX ; f <= MAXOUTPUTFILEINDEX ; f++ )
	{
	    if( OFD[ f ].rdnLinecp != NULLCP )
	    	free( OFD[ f ].rdnLinecp ) ;

	    if( rdnLinecp != NULLCP )
    		OFD[ f ].rdnLinecp = strdup( rdnLinecp ) ;
	    else
    		OFD[ f ].rdnLinecp = NULLCP ;
		
	    OFD[ f ].RDNoutput = FALSE  ;
	}
	
	return ;
}

/*
 * Attempt to open a file for writing that does not already exist.
 * Append on a counter to find a new file name if necessary e.g. "ccDataUpdata.done4".
 */

int
openAnOutputFile( file_bufferp, OFDp, tail_stringp )
char * 	file_bufferp ;
struct	outputFileDetails * OFDp ;
char *  tail_stringp ;
{
int	ret_val = NOTOK ;
char	name_buffer[ FILENAMEBUFFLEN ], tail_count_buffer[ FILENAMEBUFFLEN ] ;
char *	tailCountcp ;
int	tail_count = 0 ;
FILE **	outFilepp = &(OFDp -> outFilep) ;

	sprintf( name_buffer, "%s%s", file_bufferp, tail_stringp ) ;

	do {

	    if( tail_count > 0 )
	    	fclose( *outFilepp ) ;
	
	    if( tail_count == 1 )
	    {
	    	/* Append on the initial tail count.. */

		sprintf( tail_count_buffer, "%d", tail_count ) ;

		tailCountcp = name_buffer + strlen( name_buffer ) ;
		
		strcat( name_buffer, tail_count_buffer ) ;
	    }
	    else if( tail_count > 1 )
	    {
	    	/* Update the tail count.. */
		
		sprintf( tail_count_buffer, "%d", tail_count ) ;

		*tailCountcp = '\0' ;

		strcat( tailCountcp, tail_count_buffer ) ;
	    }
	    
	
	    /* Detemine whether a file of the current "name_buffer" file already exists.. */
	    (*outFilepp) = fopen( name_buffer, "r" ) ;
	    
	} while( ++tail_count <= MAXTAILCOUNT && (*outFilepp) != NULL ) ;

	if( (*outFilepp) != NULL )
	{
	    fprintf( stderr, "Could not derive a file name for %s%s\nbut tried upto %s\n",
	    		file_bufferp, tail_stringp, name_buffer );
			
	    fclose( (*outFilepp) ) ;
	}
	else
	{
	    /* attempt to open for writing (file create).. */
	    if( ((*outFilepp) = fopen( name_buffer, "w" ) ) != NULL )
	    {
	    	strcpy( OFDp -> outputFilename, name_buffer) ;

	    	OFDp -> RootedAtOutput = FALSE ;
		OFDp -> RootedAtDNcp   = NULLCP ;
		OFDp -> SubDirListResolvedParentDNcp = NULLCP ;
	    	
		OFDp -> RDNoutput = FALSE ;
		OFDp -> rdnLinecp = NULLCP ;
		
		ret_val = OK ;
	    }
#ifdef DEBUG
fprintf( stderr, "File name %s\n", name_buffer) ;
#endif
	}

	return( ret_val ) ;
}

int
initialiseDraftFile( OFD )
struct	outputFileDetails OFD[] ;
{
int	retVal = OK ;
char *	draftFilecp, draftFileBuff[ SMALLBUFLEN ] ;

	sprintf( draftFileBuff, "%sXXXXXX", DRAFTFILENAME ) ;
	draftFilecp = mktemp( draftFileBuff ) ;

	if( (OFD[ OUTPUTDRAFTFILE ].outFilep = fopen( draftFilecp, "w" ) ) == ((FILE *)NULL) )
	{
	    perror( "Attempt to open the bulk load draft file for writing failed" );
	    retVal = NOTOK ;
	}
	else strcpy( OFD[ OUTPUTDRAFTFILE ].outputFilename, draftFilecp ) ;
	
	
	return( retVal ) ;
}

int
initialiseOutFiles( file_buffercp, OFD )
char *  file_buffercp ;
struct	outputFileDetails OFD[] ;
{
int	ret_val = NOTOK ;

 	if( initialiseDraftFile( OFD ) == OK )
	{
	    if( openAnOutputFile( file_buffercp, &OFD[ OUTPUTDONEFILE ]    , DONEFILETAIL       ) == OK &&
	        openAnOutputFile( file_buffercp, &OFD[ OUTPUTNOACTIONFILE ], NOACTIONFILETAIL   ) == OK &&
	        openAnOutputFile( file_buffercp, &OFD[ OUTPUTPROBLEMFILE ] , PROBLEMFILETAIL    ) == OK   )
	    {
	    	resetOutFileDetails( OFD ) ;

	    	ret_val = OK ;
	    }
	}

	return( ret_val ) ;
}

/******************************************************************************/


int
determineLineType( lineBufferp )
char * lineBufferp ;
{
int lineType = UNDEFLINETYPE ;

	if( isspace( *lineBufferp ) )
	    lineType = BLANKLINETYPE ;

	else if ( *lineBufferp == COMMENTCHAR )
	{
	    if( !strncmp( lineBufferp, "#ERR", 4 ) )
	    	lineType = ERRORCOMMENTLINETYPE ;
	    else
	    	lineType = COMMENTLINETYPE ;
	}
	else if ( *lineBufferp == MODRDNCHAR )
	    lineType = MODRDNLINETYPE ;

	else if ( !strncasecmp( lineBufferp, "source", 6 ) )
	    lineType = SOURCELINETYPE ;

	else if ( !strncasecmp( lineBufferp, "rootedAt", 8 ) )
	    lineType = ROOTEDATLINETYPE ;

	else if ( strchr( lineBufferp, ATTRTYPEDELIM ) != NULL ||
		  *lineBufferp == DELETEATTRCHAR ) /* When no value will be present */
	    lineType = ATTSLINETYPE ;

	else if( !strncasecmp( lineBufferp, NEWDMFORMATFIRSTLINE, strlen( NEWDMFORMATFIRSTLINE ) ) )
	    lineType = NEWDMFORMATLINETYPE ;
	
	else lineType = OTHERLINETYPE ; /* E.g. empty string "" */


	return( lineType ) ; /* XXXLINETYPE */
}


int 
matchLineType( lineType, ReqLineType )
int lineType, ReqLineType ;
{
int	ret_val = FALSE ;

	if( ReqLineType == ANYLINETYPE )
	{
	    if( lineType != OTHERLINETYPE   &&
	    	lineType != BLANKLINETYPE   &&
	    	lineType != COMMENTLINETYPE &&
	    	lineType != ERRORCOMMENTLINETYPE )
	    	ret_val = TRUE ;
	}
	if( ReqLineType == ROOTEDATATTSORBLANKLINETYPE )
	{
	    if( lineType == ROOTEDATLINETYPE ||
  		lineType == ATTSLINETYPE     ||
		lineType == BLANKLINETYPE      )
		    ret_val = TRUE ;
	}
	if( ReqLineType == ROOTEDATMODRDNATTSORBLANKLINETYPE )
	{
	    if( lineType == ROOTEDATLINETYPE ||
	        lineType == MODRDNLINETYPE   ||
  		lineType == ATTSLINETYPE     ||
		lineType == BLANKLINETYPE      )
		    ret_val = TRUE ;
	}
	else if( ReqLineType == ROOTEDATORATTSLINETYPE )
	{
	    if( lineType == ROOTEDATLINETYPE ||
  		lineType == ATTSLINETYPE     )
		    ret_val = TRUE ;
	}
	else
	{
	    if( lineType == ReqLineType )
	    	    ret_val = TRUE ;
	}

#ifdef DEBUG
fprintf( stderr, "Line in %d Match with %d = %d\n", lineType, ReqLineType, ret_val ) ;
#endif

	return( ret_val ); /* TRUE or FALSE */
}


/*
 * Strip off the attribute type (without the equals sign !)
 *
 * e.g. telephoneNumber from "-telephoneNumber"       - newdmdiff format
 *			     ">telephoneNumber= 9999" - newdmdiff format
 *			     "telephoneNumber= 9999"  - dmformat format
 */

void
extractAttrType( returnAttrTypeBuffer, attrLineBuffer, returnAttrTypeLenp, inFileType )
char 	returnAttrTypeBuffer[], attrLineBuffer[] ;
int  *	returnAttrTypeLenp, inFileType ;
{
char *	currAttrLinecp = attrLineBuffer ;
int	i=0 ;

	if( inFileType == NEWDMFORMATINPUT )
	{
	    if( isCharInSet( *currAttrLinecp, ALLBULKLOADCHARS ) == TRUE )
	    	currAttrLinecp++ ; /* Skip leading character */
	}
#ifdef DEBUG
else if( isCharInSet( *currAttrLinecp, ALLBULKLOADCHARS ) == TRUE )
	fprintf( stderr, "Warning this line could be NEWDMFORMAT [%c] %s\n", 
			*currAttrLinecp, currAttrLinecp ) ;
#endif
	
	while( *currAttrLinecp != '\0' &&
	       !isspace( *currAttrLinecp ) &&
	       *currAttrLinecp != ATTRTYPEDELIM )
	{
	    returnAttrTypeBuffer[i] = *currAttrLinecp++ ;
	    i++ ;
	}
	
	returnAttrTypeBuffer[ i ] = '\0' ;

	*returnAttrTypeLenp = i ;
	
#ifdef DEBUG
fprintf( stderr, "Attr type = %s\n", returnAttrTypeBuffer ) ;
#endif
	return ;
}

/*
 * Line continutation requires an end of line "\".."
 */
 
int
lineToBeContinued( lineBufferp )
char *	lineBufferp ;
{
int	lineContReq = FALSE ;
char *  lineContcp ;

	if( (lineContcp = strrchr( lineBufferp, LINECONTCHAR )) != NULL &&
	    *(++lineContcp) == '\n' )
	{
	    lineContReq = TRUE ;
	}

	return( lineContReq ) ;
}

/*
 * Inherited attributes are treated as a special case
 */

int
iattrLineToBeContinued( lineBuffercp, firstLine, inFileType )
char *	lineBuffercp ;
int	firstLine, inFileType ;
{
int	lineContReq = FALSE, attrTypeLen ;
char    attrLineType[SMALLBUFLEN] ;
AttributeType	at ;
static	AttributeType iattrat = NULLAttrT ;

	if( iattrat == NULLAttrT )
	    iattrat =  str2AttrT( BL_INHERITED_ATTRIBUTE ) ;

	if( iattrat != NULLAttrT )
	{
	    if( firstLine == TRUE )
	    {
		extractAttrType( attrLineType, lineBuffercp, &attrTypeLen, inFileType ) ;
	    
		if( (at = str2AttrT( attrLineType )) != NULLAttrT )
		{
		    if( AttrT_cmp( at, iattrat ) == 0 )
			lineContReq = TRUE ;
		    
		    AttrT_free( at ) ;
		}
		
	    } else {

		/* Subsequent line so check for solitary closing brace.. */
		if( *lineBuffercp != ')' || *(lineBuffercp+1) != '\n' )
		    lineContReq = TRUE ;
	    }
	}

	return( lineContReq ) ;
}


/*
 * Attempt to find the next line of the specified type.
 *
 * Note: The line buffer may already contain an unprocessed line.
 *
 * Multiple line shall be buffered if line continuation characters
 * are found or if an inheritedAttribute is being processed.
 */

int
getLine( IFDp, lineBufferStructp, ReqLineType, inFileType )
struct	inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
int	ReqLineType, inFileType ;
{
int	ret_val  = OK ;
int	lineType = UNDEFLINETYPE ;
int	matchOK  = FALSE ;
int 	LineContinued = FALSE, IattrLineContinued = FALSE ;
char *	baseBuffercp = lineBufferStructp -> lineBufferp, 
     *	currBuffercp = lineBufferStructp -> lineBufferp ;
int	fileEOF, bufferSizeUsed = 0 ;

	/* The line buffer may already contain a line if the correct type.. */
	lineType = determineLineType( lineBufferStructp -> lineBufferp ) ;
	
	while( !(fileEOF = feof( IFDp -> parseInputFilep )) &&
 	       ret_val == OK &&
	       (  (matchOK = matchLineType( lineType, ReqLineType)) == FALSE
		  ||
	          (matchOK == TRUE && (LineContinued == TRUE || IattrLineContinued == TRUE) ))
	     )
	{
	    /* Check that the line buffer is large enough.. */

	    if( bufferSizeUsed >= (lineBufferStructp->bufferSize - MAXLINELEN))
	    {
	    	/* Request a larger buffer.. */
		if( (lineBufferStructp -> lineBufferp = 
			realloc( (char *)lineBufferStructp->lineBufferp, 
				 (unsigned)(lineBufferStructp->bufferSize + LINEBUFSIZEINC))) == NULL )
		{
		    fprintf( stderr, "Memory REALLOCATION failure when increaing buffer sizeto %d\n", 
		    		lineBufferStructp->bufferSize + LINEBUFSIZEINC );		    

		    ret_val = NOTOK ;
		}
		else 
		{
		    fprintf( stderr, "Realloc'ed to %d\n", lineBufferStructp->bufferSize + LINEBUFSIZEINC );

		    lineBufferStructp->bufferSize += LINEBUFSIZEINC ;

		    /* Adjust buffer pointer to point to the same
		       position in the realloc'ed buffer ..       */

		    currBuffercp = lineBufferStructp -> lineBufferp +
		    				(currBuffercp - baseBuffercp) ;

		    baseBuffercp = lineBufferStructp -> lineBufferp ;
		}
	    }


	    if( ret_val == OK )
	    {
		if( matchOK == FALSE )
		{
		    currBuffercp = lineBufferStructp -> lineBufferp ; /* Reset the line buffer ptr */

		    bufferSizeUsed = 0 ;
		}
	
		/* Read in a fresh line.. */
		if( fgets( currBuffercp, MAXLINELEN, IFDp -> parseInputFilep ) == NULL )
{
#ifdef DEBUG
fprintf( stderr," FILE END WAS REACHED\n" );
#endif
	    	    ret_val = NOTOKALLDATAREAD ;
}
	    	else
		{
{
#ifdef DEBUGALLLINES
fprintf( stderr, ">>%s", currBuffercp ) ;
#endif
}		    
		    if( LineContinued == FALSE && IattrLineContinued == FALSE )
		    {
			/* ..no line continuations so far */
		    
	    	    	lineType = determineLineType( baseBuffercp ) ;

			/* Transfer the file to the parse output file.. */
			if( lineType != ERRORCOMMENTLINETYPE )
			    fprintf( IFDp -> parseOutputFilep, "%s", currBuffercp ) ;

		     	/* Check if this line should be continued.. */
		     	if( lineToBeContinued( currBuffercp ) == TRUE )
		     	{
			    LineContinued  =  TRUE ;
			    bufferSizeUsed += strlen( currBuffercp ) ;
			    currBuffercp   =  baseBuffercp + bufferSizeUsed ;
			}

			/* Check if this line contains an iattr attribute.. */
			if( lineType == ATTSLINETYPE && inFileType != UNDEFFORMATINPUT &&
			    (IattrLineContinued = iattrLineToBeContinued( currBuffercp, TRUE, inFileType )) == TRUE )
			{
			    bufferSizeUsed += strlen( currBuffercp ) ;
			    currBuffercp   =  baseBuffercp + bufferSizeUsed ;
			}
		    }
		    else
		    {
		    	/* handle line continuation .. */

			if( LineContinued == TRUE )
			{
		     	    if( lineToBeContinued( currBuffercp ) == FALSE )
			    	LineContinued = FALSE ; /* Continuation finished */
			    else
			    {
			    	/* Prepare for appending the next line */
			    	bufferSizeUsed += strlen( currBuffercp ) ;
			    	currBuffercp   =  baseBuffercp + bufferSizeUsed ;
			    }
			}
			else if( IattrLineContinued == TRUE )
			{
		     	    if( iattrLineToBeContinued( currBuffercp, FALSE, inFileType ) == FALSE )
			    	IattrLineContinued = FALSE ; /* Continuation finished */
			    else
			    {
			    	/* Prepare for appending the next line */
			    	bufferSizeUsed += strlen( currBuffercp ) ;
			    	currBuffercp   =  baseBuffercp + bufferSizeUsed ;
			    }
			}
		    }
		}
	    }
	    
	} /* While loop */

	if( fileEOF || ( matchOK == FALSE && ret_val != NOTOK ) )
	    ret_val =  NOTOKALLDATAREAD ;

	return( ret_val ); /* OK, NOTOKALLDATAREAD */
}


RDN
extractRDN( rdnLinecp ) 
char *	rdnLinecp ;
{
RDN	retRDN = NULLRDN ;


	/* Check for a leading bulk load character.. */
	if( *rdnLinecp == ADDRDNCHAR || *rdnLinecp == DELETERDNCHAR )
	    rdnLinecp++ ; /* skip it */
	
	/* Check for trailing '\n' in the RDN line.. */
	if( *(rdnLinecp + strlen( rdnLinecp ) - 1 ) == '\n' )
	    *(rdnLinecp + strlen( rdnLinecp ) - 1 ) =  '\0' ;

	rdnLinecp = TidyString( rdnLinecp ) ;

	if( (retRDN = str2bulkRdn( rdnLinecp ) ) == NULLRDN )
	{
	    fprintf( stderr, "RDN parse failure for %s\n", rdnLinecp ) ;
	}
	
	return( retRDN ) ;
}


/*****************************************************************************/

void
free_attrLine_list( baseAttrLinep )
struct attributeLine * baseAttrLinep ;
{
struct attributeLine * currAttrLinep = baseAttrLinep,
		     * buffAttrLinep = NULL ;

	while( currAttrLinep != NULL )
	{
	    /* Free the attribute character line storage.. */
	    (void)free( (char*)(currAttrLinep -> attrLinecp) ) ;

	    if( currAttrLinep -> errorMsgcp != NULLCP )
	    	free( currAttrLinep -> errorMsgcp ) ;
	    
	    if( currAttrLinep -> at != NULLAttrT )
	    	AttrT_free( currAttrLinep -> at ) ;

	    if( currAttrLinep -> av != NULLAttrV )
	    	AttrV_free( currAttrLinep -> av ) ;
	    
	    buffAttrLinep = currAttrLinep ;
	    currAttrLinep = currAttrLinep -> nextAttrLinep ;

	    /* Free the node storage.. */
	    (void)free( (char*)( buffAttrLinep )) ;
	}

	return ;
}




/*
 * Free the entire "rootedAtAttributes" linked list..
 */
 
void
freeRootedAtAttrs_list( baseRAApp )
struct	rootedAtAttributes ** baseRAApp ;
{
struct 	rootedAtAttributes * currRAAp = *baseRAApp,
		           * buffRAAp = NULL ;

	while( currRAAp != NULL )
	{
	    /* Free the RDN character line storage.. */
	    if( currRAAp -> rdnLinecp != NULLCP )
		(void)free((char*)( currRAAp -> rdnLinecp) ) ;

	    /* Free the RDN storage.. */
	    if( currRAAp -> rdn != NULLRDN )
		rdn_free( currRAAp -> rdn ) ;

	    if( currRAAp -> modRdn != NULLRDN )
	    	rdn_free( currRAAp -> modRdn ) ;
	    
	    /* Free the rootedAt DN storage.. */
	    if( currRAAp -> rootedAtDN != NULLDN )
		dn_free( currRAAp -> rootedAtDN ) ; 

	    /* Free the attribute sequence error message.. */
	    if( currRAAp -> attrSequenceErrorMsgcp != NULLCP )
	    	free( currRAAp ->  attrSequenceErrorMsgcp ) ;
	    
	    /* Free the attributeLine storage.. */
	    free_attrLine_list( currRAAp -> attrLinep ) ;

	    buffRAAp = currRAAp ;
	    currRAAp = currRAAp -> nextRootedAtAttrsp ;

	    /* Free the node storage.. */
	    (void)free((char*)( buffRAAp )) ;
	}

	/* All nodes returned.. */
	*baseRAApp = NULLRAA ;

	return ;
}


int
setUpTypeValueFromLine( IFDp, ALp, attrLinecp, inFileType )
struct  inFileDetails * IFDp ;
struct	attributeLine * ALp ;
char *	attrLinecp  ;
int	inFileType ;
{
int	ret_val = OK ;
char	attrLineType[SMALLBUFLEN], * valuecp, * cp, * atcp = attrLineType ;
int	attrLineLen, fullStopFound = FALSE ;
AttributeType  at ;
static	AttributeType cnat = NULLAttrT, iattrat = NULLAttrT ;

	ALp -> at = NULLAttrT ; ALp -> av = NULLAttrV ;

	if( cnat == NULLAttrT )
	    cnat =  str2AttrT( BL_COMMON_NAME ) ;

	if( iattrat == NULLAttrT )
	    iattrat =  str2AttrT( BL_INHERITED_ATTRIBUTE ) ;


	extractAttrType( atcp, ALp -> attrLinecp, &attrLineLen, inFileType ) ;

	at = str2AttrT( atcp ) ;

	if( at != NULLAttrT )
	{
	    ALp -> at = at ;
	    	
	    if( at -> oa_syntax == 0 )
		reportParseError( IFDp, "Unknown attribute type syntax for -", atcp ) ;
		
	    else if( (valuecp = strchr( attrLinecp, '=' ) ) == NULL )
	    {
	    	if( inFileType != NEWDMFORMATINPUT || *attrLinecp != DELETEATTRCHAR )
		    reportParseError( IFDp, "Missing '=' within -", attrLinecp ) ;
	    }
	    else{

		if( iattrat != NULLAttrT && AttrT_cmp( iattrat, at ) == 0 )
		{
		    while( isspace( *++valuecp ) )
		    	/* Skip leading whitespace for multi-line inherited attributes */ ;

		    valuecp = strdup( valuecp ) ;
		}
		else
		{
		    valuecp = strDupTidyStringRemLineConts( ++valuecp ) ;
		}
	    
		if( valuecp != NULL )
		{ 
		    /* ...skipping '=' */

		    /* If commonName replace any full stop delimiters with spaces.. */
		    if( bulkLoadConfig.replaceFullStopCommonNameDelimsWithSpaces == TRUE &&
		        AttrT_cmp( cnat, at ) == 0 )
		    {
		    	cp = valuecp ;

			while( *cp )
			{
			    if( *cp == '.' )
			    {
				fullStopFound = TRUE ;
			    
				*cp++ = ' ' ;
			    }
			    else cp++ ;
			}

			if( fullStopFound == TRUE )
			    valuecp = TidyString( valuecp ) ; /* In case any double-spaces have been created */
		    }

	    	    ALp -> av = str2AttrV( valuecp, ALp -> at -> oa_syntax) ;

		    if( ALp -> av == NULLAttrV )
			reportParseError( IFDp, "str2AttrV could not encode -", valuecp ) ;

		    free( valuecp ) ;
		}
	    }

	    if( ALp -> av == NULLAttrV && 
	        (inFileType != NEWDMFORMATINPUT || *attrLinecp != DELETEATTRCHAR ))
	    {
		AttrT_free( ALp -> at ) ;
		ALp -> at = NULLAttrT ;
#ifdef DEBUG
fprintf( stderr, ">VALUE FAILURE %s \nmsyntax%d\nvaluecp=%s\n", attrLinecp, at ->oa_syntax, valuecp ) ;
#endif
		ret_val = NOTOK ;
	    }
	}
	else 
	{
	    reportParseError( IFDp, "str2AttrT() could not encode -", atcp ) ;

	    ret_val = NOTOK ;
	}
	
	return( ret_val ) ;
}


int
setUpTypeValue( IFDp, ALp, atcp, avcp )
struct	inFileDetails * IFDp ;
struct	attributeLine * ALp ;
char *	atcp, * avcp ;
{
int	ret_val = OK, fullStopFound = FALSE ;
char *	valuecp, * cp ;
AttributeType  at ;
static	AttributeType cnat = NULLAttrT ;

	ALp -> at = NULLAttrT ; ALp -> av = NULLAttrV ;

	if( cnat == NULLAttrT )
	    cnat =  str2AttrT( BL_COMMON_NAME ) ;

	if( atcp == NULL || avcp == NULL )
	    return( NOTOK ) ;

	at = str2AttrT( atcp ) ;

	if( at != NULLAttrT )
	{
	    ALp -> at = at ;
	    	
	    if( at ->oa_syntax == 0 )
		reportParseError( IFDp, "Unknown attribute type syntax for -", atcp ) ;
	    
	    else {
	    
		if(( valuecp = strDupTidyStringRemLineConts( avcp )) != NULL ) 
		{ 
		    /* ...skipping '=' */

		    /* If commonName replace any full stop delimiters with spaces.. */
		    if( bulkLoadConfig.replaceFullStopCommonNameDelimsWithSpaces == TRUE &&
		        AttrT_cmp( cnat, at ) == 0 )
		    {
		    	cp = valuecp ;

			while( *cp )
			{
			    if( *cp == '.' )
			    {
				fullStopFound = TRUE ;
			    
				*cp++ = ' ' ;
			    }
			    else cp++ ;
			}

			if( fullStopFound == TRUE )
			    cp = TidyString( cp ) ; /* In case any double-spaces have been created */
		    }

	    	    ALp -> av = str2AttrV( valuecp, ALp -> at -> oa_syntax) ;

		    if( ALp -> av == NULLAttrV )
			reportParseError( IFDp, "str2AttrV could not encode -", valuecp ) ;

		    free( valuecp ) ;
		}		    
	    }

	    if( ALp -> av == NULLAttrV )
	    {
		AttrT_free( ALp -> at ) ;
		ALp -> at = NULLAttrT ;

#ifdef DEBUG
fprintf( stderr, "VALUE FAILURE %s\n", avcp ) ;
#endif
		ret_val = NOTOK ;
	    }
	}
	else 
	{
	    ret_val = NOTOK ;

	    reportParseError( IFDp, "str2AttrT() could not encode -", atcp ) ;
#ifdef DEBUG
fprintf( stderr, "TYPE FAILURE %s\n", atcp ) ;
#endif
	}

	return( ret_val ) ;
}


char	returnString[2] = { "\n" } ;

struct	attributeLine * 
create_attributeLine_node_from_type_value( IFDp, leadingChar, attrTypecp, attrValuecp )
struct  inFileDetails    * IFDp ;
char	leadingChar ; /* 0 if none */
char *	attrTypecp, * attrValuecp ;
{
struct 	attributeLine * newAttributeLinep = NULL ;
char *  attrLinecp = NULL ;
int	returnStorage = FALSE ;

	if( attrTypecp == NULLCP || attrValuecp == NULLCP )
	    return( NULL ) ;

	if( (newAttributeLinep = (struct attributeLine *)malloc( 
				(unsigned)sizeof( struct attributeLine )) ) != NULLALP )
	{
	    /* ..single attribute value */

	    /* Initialise pointers to their default settings incase a failure occurs.. */

	    newAttributeLinep -> attrLinecp = NULLCP ;
	    newAttributeLinep -> at 	    = NULLAttrT ;
	    newAttributeLinep -> av 	    = NULLAttrV ;
	    newAttributeLinep -> nextAttrLinep = NULLALP ;


	    attrLinecp = malloc( (unsigned)(strlen( attrTypecp ) + strlen( attrValuecp ) + 5)) ;

	    if( attrLinecp )
	    {
		/* Load up the attribute type attribute value pair with any leading newdmformat
		   character and the required trailing '\n'.. */

		newAttributeLinep -> attrLinecp = attrLinecp ;

		if( leadingChar )
		    sprintf( newAttributeLinep -> attrLinecp, "%c%s%c%s",
				 leadingChar, attrTypecp, ATTRTYPEDELIM, attrValuecp ) ;
		else sprintf( newAttributeLinep -> attrLinecp, "%s%c%s",
				 attrTypecp, ATTRTYPEDELIM, attrValuecp ) ; /* No leading character */

		if( *( attrValuecp + (strlen(attrValuecp) - 1)) != '\n' )
		    strcat( newAttributeLinep -> attrLinecp, returnString ) ;

		newAttributeLinep -> attrUsed = FALSE ;

		newAttributeLinep -> attrIsMasteredByCurrentSource = FALSE ;

		newAttributeLinep -> errorMsgcp = NULLCP ;

		newAttributeLinep -> actionDefinedByAttrLine = 0 ;

		newAttributeLinep -> modificationInitiated = 0 ;
		
		newAttributeLinep -> nthAttrOfThisType = 0 ;
#ifdef DEBUG
fprintf( stderr, "Adding attr [%s]", newAttributeLinep -> attrLinecp ) ;
#endif
	    	if( setUpTypeValue( IFDp, newAttributeLinep, attrTypecp, attrValuecp ) == NOTOK )
		    returnStorage = TRUE ;
	    }
	    else
	    {
		reportFatalParseError( IFDp, "Memory allocation failure when creating an attributeLine", NULLCP ) ;
#ifdef DEBUG
fprintf( stderr, "Memory allocation failure when creating an attributeLine\n" );
#endif
	    	returnStorage = TRUE ;
	    }
	}
	else
	{
	    reportFatalParseError( IFDp, "Memory allocation failure when creating an attributeLine", NULLCP ) ;

	    /* Note: no storage to return */
#ifdef DEBUG
fprintf( stderr, ">Memory allocation failure when creating an attributeLine node\n" );
#endif
	}

	if( returnStorage == TRUE )
	{
	    free_attrLine_list( newAttributeLinep ) ;

	    newAttributeLinep = NULL ;
	}

	return( newAttributeLinep ) ;
}


void
skipTailLineCont( attrValuecpp )
char **	attrValuecpp ;
{
char *  attrValuecp = *attrValuecpp ;

	/* Check to see if there is broken line character
	   before the next non whitespace character */

	while( attrValuecp && isspace(*attrValuecp) )
	{
	    attrValuecp++ ;
	}

	if( *attrValuecp == LINECONTCHAR )
	{
	    /* skip Line continuation character */
	    attrValuecp++ ;

	    /* Skip upto the new line character */
	    while( *attrValuecp && 
	    	   (*attrValuecp == ' ' || *attrValuecp == '\t') )
	    {
	    	attrValuecp++ ;
	    }

	    /* Load up the return pointer */
	    if( *attrValuecp == '\n' )
	    	*attrValuecpp = attrValuecp + 1 ; /* Skipping the `\n` */
	}

	return ;
}

struct	attributeLine *
create_lone_attributeLine_node( leadNewDMFormatChar, at, av )
char	leadNewDMFormatChar ;
AttributeType	at ;
AttributeValue	av ;
{
struct	attributeLine * newALp = NULLALP ;
char *	attrLinecp = NULLCP, * atcp = NULLCP, * avcp = NULLCP ;
char	leadNewDMFormatCharStr[ 2 ] ;

	leadNewDMFormatCharStr[ 1 ] = '\0' ;
	
	if( at != NULLAttrT)
	    atcp = AttrT2pstr( at ) ;

	if( av != NULLAttrV)
	    avcp = AttrV2pstr( av ) ;

	if( atcp != NULLCP &&
	    (av == NULLAttrV || avcp != NULLCP) &&
	    (attrLinecp = malloc( strlen( atcp ) + ((avcp != NULLCP) ? strlen( avcp ) : 0) + 5 )) != NULLCP &&
	    (newALp = (struct attributeLine *)malloc( (unsigned)sizeof( struct attributeLine ) )) != NULLALP  )
	{
	    /* Form the attribute line.. */
	    newALp -> attrLinecp = attrLinecp ;

	    leadNewDMFormatCharStr[ 0 ] = leadNewDMFormatChar ; /* In case we are not delaing with a NEWDMFORMAT file */

	    if( avcp != NULLCP )
	    	sprintf( newALp -> attrLinecp, "%s%s%s%s\n", leadNewDMFormatCharStr, atcp, "= ", avcp ) ;
	    else
	    	sprintf( newALp -> attrLinecp, "%s%s\n", leadNewDMFormatCharStr, atcp ) ;	    

	    newALp -> at = AttrT_cpy( at ) ;
	    newALp -> av = (av == NULLAttrV) ? NULLAttrV : AttrV_cpy( av ) ;

	    newALp -> attrUsed = FALSE ;
	    newALp -> attrIsMasteredByCurrentSource = FALSE ;
	    newALp -> errorMsgcp = NULLCP ;
	    newALp -> actionDefinedByAttrLine = 0 ;
	    newALp -> modificationInitiated   = 0 ;
	    newALp -> nthAttrOfThisType = 0 ;

	    newALp -> nextAttrLinep = NULLALP ;
	}
	else if( attrLinecp != NULLCP )
	    free( attrLinecp ) ;

	if( atcp != NULLCP ) free( atcp ) ;
	if( avcp != NULLCP ) free( avcp ) ;

	return( newALp ) ;
}

/*
 * Create an "attributeLine" node to store the attribute type value pair.
 * The procedure must handle multi-value attributes e.g. "telephoneNumber = 2345 & 3233"
 * aswell as attribute values that are broken across lines by the "\" operator.
 *
 * The second parameter will be loaded with the address of "nextAttrLinep"
 * in the final created node.
 */

struct attributeLine *
create_attributeLine_node( IFDp, attrLinecp, finalNextAttrLineppp, inFileType )
struct  inFileDetails    * IFDp ;
char  * attrLinecp ;
struct 	attributeLine *** finalNextAttrLineppp ;
int	inFileType ;
{
struct 	attributeLine * newAttributeLinep = NULL, * currAttributeLinep = NULL ;
char  *	attrValcp  = NULLCP ;
char  * attrTypecp = NULLCP ;
char  * nextAttrValuecp = attrLinecp ;
char	bufferedAttrType[SMALLBUFLEN], leadingChar = 0 ;
int	bufferedAttrTypeLen  ;
int	AllAttrValuesProcessed = FALSE, multiAttrValues = FALSE ;
int 	returnStorage ;

	do {

	    returnStorage = FALSE ;

	    /* Check to see if the line contains multiple attribute values */

	    if( (attrValcp  = strchr( attrLinecp, MULTIATTRVALSCHAR)) != NULLCP &&
		(attrTypecp = strchr( attrLinecp, ATTRTYPEDELIM    )) != NULLCP   )
	    {
		if(inFileType == NEWDMFORMATINPUT &&
		   isCharInSet( *attrLinecp, ALLBULKLOADCHARS ) == TRUE )
		    leadingChar = *attrLinecp ;
		
	    	extractAttrType( bufferedAttrType, attrLinecp, &bufferedAttrTypeLen, inFileType ) ;

		nextAttrValuecp = attrValcp + 1 ; /* skip the MULTIATTRVALUECHAR */
		
		skipTailLineCont( &nextAttrValuecp ) ;

		*attrValcp = '\0' ;	/* Null terminate the "attrType = attrValue" */

		attrValcp  = attrTypecp + 1 ;	  /* Skip the "=" */

		multiAttrValues    = TRUE ;

	    	currAttributeLinep = newAttributeLinep = 
				create_attributeLine_node_from_type_value( IFDp,
					leadingChar, bufferedAttrType, attrValcp ) ;
	    }
	    else if( multiAttrValues == FALSE )
	    {
	    	if( ( currAttributeLinep = newAttributeLinep = 
			(struct attributeLine *)malloc( 
					(unsigned)sizeof( struct attributeLine )) )!= NULLALP )
	    	{
		    /* ..single attribute value */

	    	    /* Initialise pointers to their default settings incase a failure occurs.. */

	    	    newAttributeLinep -> attrLinecp = NULLCP ;
	    	    newAttributeLinep -> at 	    = NULLAttrT ;
	    	    newAttributeLinep -> av 	    = NULLAttrV ;
	    	    newAttributeLinep -> nextAttrLinep = NULLALP ;

		    /* Remove additional leading white spaces.. IS THIS NECESSARY ? */
/*	    	    while( *attrLinecp && isspace( *(attrLinecp + 1) ) )
			attrLinecp++ ;*/

		    if( isspace( *attrLinecp ) )
			*attrLinecp = ' ' ; /* Ensure the leading white space is a ` ` */

		    newAttributeLinep -> attrLinecp = strdup( attrLinecp ) ;
#ifdef DEBUG
fprintf( stderr, "Adding attr %s", attrLinecp ) ;
#endif
		    newAttributeLinep -> attrUsed = FALSE ;

		    newAttributeLinep -> attrIsMasteredByCurrentSource = FALSE ;

		    newAttributeLinep -> errorMsgcp = NULLCP ;

		    newAttributeLinep -> actionDefinedByAttrLine = 0 ;

		    newAttributeLinep -> modificationInitiated = 0 ;

		    newAttributeLinep -> nthAttrOfThisType = 0 ;

		    if( setUpTypeValueFromLine( IFDp, newAttributeLinep, attrLinecp, inFileType ) == NOTOK )
			returnStorage = TRUE ;
		}
		else
		{
	    	    reportFatalParseError( IFDp, "Memory allocation failure when creating an attributeLine", NULLCP ) ;
		}

		if( returnStorage == TRUE )
		{
		    free_attrLine_list( newAttributeLinep ) ;
		    currAttributeLinep = newAttributeLinep = NULLALP ;
		}

		AllAttrValuesProcessed = TRUE ;
	    }
	    else /* multiAttrValues == TRUE */
	    {
	    	if( (attrValcp  = strchr( nextAttrValuecp, MULTIATTRVALSCHAR)) != NULL )
		{
	    	    /* Add the next attribute value to the linked list.. */

		    *attrValcp = '\0' ; /* Null terminate the attribute value */

		    if( (currAttributeLinep -> nextAttrLinep =
		    	create_attributeLine_node_from_type_value( IFDp,
			    leadingChar, bufferedAttrType, nextAttrValuecp ) ) == NULL )
		    	AllAttrValuesProcessed = TRUE ;
		    else
		    {
		    	nextAttrValuecp = attrValcp + 1 ; /*Skip the position the held
		    					    the "&" before being replaced
							    by the null terminator */

			skipTailLineCont( &nextAttrValuecp ) ;

			currAttributeLinep = currAttributeLinep -> nextAttrLinep ;		    		
		    }
		}
		else 
		{
	    	    /* Add the final attribute value to the linked list.. */

		    currAttributeLinep -> nextAttrLinep =
			create_attributeLine_node_from_type_value( IFDp,
			    leadingChar, bufferedAttrType, nextAttrValuecp ) ;

		    currAttributeLinep = currAttributeLinep -> nextAttrLinep ;

		    AllAttrValuesProcessed = TRUE ;
		}
	    }
	
	} while( AllAttrValuesProcessed == FALSE &&
		 newAttributeLinep != NULL ) ;


	if( newAttributeLinep == NULLALP )
	    (*finalNextAttrLineppp) = NULL ;
	else
	    (*finalNextAttrLineppp) = &(currAttributeLinep -> nextAttrLinep) ;

	
	return( newAttributeLinep ) ; /* ptr may be NULL */

} /* create_attributeLine_node() */


struct rootedAtAttributes *
create_rootedAtAttrs_node( IFDp, RDNLinecp, rootedAtDNcp )
struct  inFileDetails    * IFDp ;
char  * RDNLinecp, * rootedAtDNcp ;
{
struct 	rootedAtAttributes * newRootedAtAttrsp = NULLRAA ;

	if( ( newRootedAtAttrsp = (struct rootedAtAttributes *)malloc( 
					(unsigned)sizeof( struct rootedAtAttributes )) )!= NULLRAA )
	{
	    newRootedAtAttrsp -> rdnLinecp = strdup( RDNLinecp ) ;

	    newRootedAtAttrsp -> rdn = extractRDN( RDNLinecp ) ;

	    newRootedAtAttrsp -> modRdn = NULLRDN ;
	    
	    newRootedAtAttrsp -> rootedAtDN = str2dn( rootedAtDNcp ) ;
#ifdef DEBUG
fprintf( stderr, "Creating rootedAttAttributes struct %s", RDNLinecp );
#endif
	    newRootedAtAttrsp -> attrSequenceErrorMsgcp = NULLCP ;
	    
	    newRootedAtAttrsp -> attrLinep = NULLALP ;

	    newRootedAtAttrsp -> nextRootedAtAttrsp = NULLRAA ;
	}

	if( newRootedAtAttrsp == NULLRAA )
	{
	    fprintf( stderr, "Memory allocation failure when creating a rootedAtAttributes node\n" );
	    reportFatalParseError( IFDp, "Memory allocation failure for rootedAtAttributes node", NULLCP ) ;
	}
	else if( newRootedAtAttrsp -> rdnLinecp  == NULLCP  ||
		 newRootedAtAttrsp -> rdn        == NULLRDN ||
		 newRootedAtAttrsp -> rootedAtDN == NULLDN    )
	{
	    if( newRootedAtAttrsp -> rdnLinecp == NULLCP )
	    	reportFatalParseError( IFDp, "Memory allocation failure when duplicating", RDNLinecp ) ;

	    else if( newRootedAtAttrsp -> rdn == NULLRDN )
	    	reportParseError( IFDp, "Could not parse the RDN -", RDNLinecp ) ;

	    else if( newRootedAtAttrsp -> rootedAtDN == NULLDN )
	    	reportParseError( IFDp, "Could not parse the rootedAt DN -", rootedAtDNcp ) ;
	
	    freeRootedAtAttrs_list( &newRootedAtAttrsp ) ;
	}
	
	return( newRootedAtAttrsp ) ; /* ptr may be NULL */

} /* create_rootedAtAttrs_node() */


/*
 * Given that the attribute type/values could have been input in any order
 * this procedure is reqired to group attribute lines involving the same
 * attribute types in to sequences.
 */
 
void
orderTheAttrsByAttrT( entryRDN, baseALpp )
RDN	entryRDN ;
struct	attributeLine ** baseALpp ;
{
struct	attributeLine * currBaseALp, * searchALp, * buffLinkNextALp, * buffUnlinkNextALp,
		      * prevALp, * nextBaseALp, * lastALpSoFarForCurrAttrT ;
int	ALisSequential, attrTypeMatchCount ;

	if( !baseALpp || *baseALpp == NULLALP )
	    return ;

	currBaseALp = *baseALpp ;


	/* Note: this loop will not cause the first attribute line node 
	   as pointed to by *baseALpp to be repositioned */

	while( currBaseALp != NULLALP )
	{
	    /* Check to see if any other attribute lines lower down in
	       the list (that are not consecutive) have the same attribute type.. */

            /* The current attribute type may be used already in the RDN.. */
	    if( findAttrTypeRdnComp( currBaseALp -> at, entryRDN ) == NULLRDN )
	   	currBaseALp -> nthAttrOfThisType = attrTypeMatchCount = 1 ;
	    else
	        currBaseALp -> nthAttrOfThisType = attrTypeMatchCount = 2 ;

	    searchALp = currBaseALp -> nextAttrLinep ;

	    ALisSequential = TRUE ; prevALp = NULLALP ; nextBaseALp = NULLALP ;

	    lastALpSoFarForCurrAttrT = currBaseALp ;

	    while( searchALp != NULLALP )
	    {
		if( AttrT_cmp( currBaseALp -> at, searchALp -> at ) == 0 )
		{
		    searchALp -> nthAttrOfThisType = (++attrTypeMatchCount) ;
		
		    if( ALisSequential == FALSE )
		    {
		    	/* Link in the current AL node with the other attributes 
			   of this attribute type.. */

			buffLinkNextALp   = lastALpSoFarForCurrAttrT -> nextAttrLinep ;
			buffUnlinkNextALp = searchALp  -> nextAttrLinep ;
			
			lastALpSoFarForCurrAttrT -> nextAttrLinep = searchALp ;
			searchALp   -> nextAttrLinep = buffLinkNextALp ;
			prevALp     -> nextAttrLinep = buffUnlinkNextALp ;
		    }

		    lastALpSoFarForCurrAttrT = searchALp ;
		}
		else
		{
		    /* Check if this is the first attribute line of a different type */
		    if( ALisSequential == TRUE )
		    {
		    	/* Indicate that this node is to be the next node to be searched from */
			nextBaseALp = searchALp ;
		    }

		    ALisSequential = FALSE ;
		}

	    	prevALp   = searchALp ;
		searchALp = searchALp -> nextAttrLinep ;
	    }

	    currBaseALp = nextBaseALp ;
	}


	/* We now run through the loop again moving add "objectClass" nodes to the base
	   and delete "objectClass" nodes to the end of the linked list */

	 

#ifdef DEBUG
searchALp = *baseALpp ;

while( searchALp != NULLALP )
{
	fprintf( stderr, "%d ATTRIBUTE LINE %s\n", searchALp -> nthAttrOfThisType, searchALp -> attrLinecp ) ;

	searchALp = searchALp -> nextAttrLinep ;
}
#endif

	return ;
}


/*
 * Buffer the attribute lines for the current RDN
 */
 
int
bufferAttrs( IFDp, lineBufferStructp, entryRDN, baseALpp, inFileType, findLineTypes )
struct  inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
RDN	entryRDN ;
struct	attributeLine ** baseALpp ;
int	inFileType, findLineTypes ;
{
struct	attributeLine ** currALpp = NULL, ** buffNewCurrALpp = NULL ;
int	AllAttrsRead = FALSE, ret_val = OK, res, atleastOneAttrRead = FALSE ;

	while( AllAttrsRead == FALSE && ret_val == OK && IFDp -> fatalErrorOccurred == FALSE )

	{
    	    if( (res = getLine( IFDp, lineBufferStructp, findLineTypes, inFileType )) == OK )
    	    {
	    	if( determineLineType( lineBufferStructp -> lineBufferp ) == ATTSLINETYPE )
		{
		    atleastOneAttrRead = TRUE ;
		
		    if( (*baseALpp) == NULL )
		    {
		    	(*baseALpp) = create_attributeLine_node( IFDp, lineBufferStructp->lineBufferp,
									&currALpp, inFileType ) ;
		    }
		    else
		    {
		    	(*currALpp) = create_attributeLine_node( IFDp, lineBufferStructp -> lineBufferp, 
										&buffNewCurrALpp, inFileType ) ;

			if( buffNewCurrALpp )
		    	    currALpp = buffNewCurrALpp ;
		    }

		    *(lineBufferStructp -> lineBufferp) = '\0' ; /* line has now been used */
		}
		else
		{
		    orderTheAttrsByAttrT( entryRDN, baseALpp ) ;
		    AllAttrsRead = TRUE ;
		}
	    }
	    else
	    {
	    	/* Check if any data has been processed.. */
	    	if( res = NOTOKALLDATAREAD )
		{
		    AllAttrsRead = TRUE ;

		    if( atleastOneAttrRead == FALSE )
		    	ret_val = res ;
		    else
		    	orderTheAttrsByAttrT( entryRDN, baseALpp ) ;
		}
		else
		    ret_val = res ; /* NOTOK - fatal error */
	    }
	    
	} /* while */
	
	return( ret_val ) ; /* OK, NOTOKALLDATAREAD or NOTOK */

} /* bufferAttrs() */

void
transferRemainingLinesBetweenParseFiles( IFDp, lineBufferStructp )
struct  inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
{
	*( lineBufferStructp -> lineBufferp ) = '\0' ; /* Line has been transferred to parse file */

	while( fgets( lineBufferStructp -> lineBufferp, MAXLINELEN, IFDp -> parseInputFilep ) != NULL )
	{
	    fprintf( IFDp -> parseOutputFilep, "%s", lineBufferStructp -> lineBufferp ) ;

	    *( lineBufferStructp -> lineBufferp ) = '\0' ; /* Line has been used */
	}
	
	return ;
}


/*
 * RDNs must have at least one blank line preceding them.
 * A single RDN modification request may immediately the RDN follow for NEWDMFORMAT files,
 * A list of attributes will then follow directly after.
 *
 * Note: "lineBuffer" may already have a line present.
 */
 
int
bufferRDNplusAttrs( IFDp, lineBufferStructp, currRAApp, inFileType, rootedAtDNcp )
struct  inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
struct  rootedAtAttributes ** currRAApp ;
int	inFileType ;
char *	rootedAtDNcp ;
{
int	getLineOK = OK, lineType = UNDEFLINETYPE,
	ret_val   = OK, someDataBuffered = FALSE ;

	ret_val = getLineOK = getLine( IFDp, lineBufferStructp, ROOTEDATATTSORBLANKLINETYPE, inFileType ) ;

	if( getLineOK == OK )
	{
	    lineType = determineLineType( lineBufferStructp -> lineBufferp ) ;

	    if( lineType == BLANKLINETYPE )
	    {
	    	/* Discard any further sequential blank lines */
		while( getLineOK == OK && lineType == BLANKLINETYPE )
		{
		    *(lineBufferStructp -> lineBufferp ) = '\0' ; /* Discard blank line */

		    ret_val = getLineOK = getLine( IFDp, lineBufferStructp, 
							ROOTEDATATTSORBLANKLINETYPE, inFileType ) ;

		    lineType = determineLineType( lineBufferStructp -> lineBufferp ) ;
		}

		if( getLineOK == OK  )
		{
		    if( getLineOK == OK && lineType == ATTSLINETYPE ) /* Should be the RDN */
	    	    {
		    	/*
			 * Buffer the RDN
			 */
			 
	    		/* Create a node to hold the retrieved RDN */
	    		(*currRAApp) = create_rootedAtAttrs_node( IFDp,
 					lineBufferStructp -> lineBufferp, rootedAtDNcp ) ;

	    		*(lineBufferStructp -> lineBufferp) = '\0' ; /* line has now been used */

	    		if( !(*currRAApp) )
	    		{
			    ret_val = NOTOK ;
	    		}
	    		else
	    		{
			    /*
			     * Check for an RDN modification request
			     */
			 
			    getLineOK = getLine( IFDp, lineBufferStructp, 
								ROOTEDATMODRDNATTSORBLANKLINETYPE, inFileType ) ;

			    if( getLineOK == OK )
			    {
			    	if( (lineType = determineLineType( lineBufferStructp -> lineBufferp )) == MODRDNLINETYPE )
				{
		    		    /* The line after the RDN is an RDN modifcation.. */
				    (*currRAApp) -> modRdn = extractRDN( lineBufferStructp -> lineBufferp + 1 ) ; /* Skipping the modify rdn character */
				    *(lineBufferStructp -> lineBufferp) = '\0' ; /* line has now been used */

				    someDataBuffered = TRUE ;
				}
			
				/* Now load up the attributes to go with the RDN.. */
			    	ret_val = bufferAttrs( IFDp, lineBufferStructp, (*currRAApp) -> rdn,
			    			&((*currRAApp) -> attrLinep), inFileType, ROOTEDATATTSORBLANKLINETYPE ) ;

			    	/* If no attributes were found then check for
			      	   the presence of an RDN modification or a delete entry request .. */

			    	if( ret_val == NOTOKALLDATAREAD &&
				    (someDataBuffered == TRUE || 
				    	((*currRAApp) -> rdnLinecp != NULLCP && *((*currRAApp) -> rdnLinecp) == DELETERDNCHAR )))
			    	    ret_val = OK ;
			    }
			    else
			    {
			    	/* Since the attempt to buffer the line after the RDN failed
				   we must check for a delete RDN request.. */
				if( inFileType != NEWDMFORMATINPUT || *((*currRAApp)->rdnLinecp) != DELETERDNCHAR )
				{
			    	    reportParseError( IFDp, "No attribute lines listed after RDN", NULLCP ) ;
			    	    ret_val = NOTOKALLDATAREAD ;
				}
			    }
			}
		    }
		}
	    }
	    else
	    {
		if( IFDp -> fileUsage != PARSEINPUTFILEINPARSE1OUT ) /* processing the results of an edit */
		{
	    	    reportParseError( IFDp, "Missing blank line before RDN", NULLCP ) ;
		    transferRemainingLinesBetweenParseFiles( IFDp, lineBufferStructp ) ;
		}

	    	ret_val  =  NOTOKALLDATAREAD ; /* no RDN has been found */
	    }
        }
	
	return( ret_val ) ; /* OK, NOTOKALLDATAREAD or NOTOK */

} /* bufferRDNplusAttrs() */



/**********************************************************************************/

/* 
 * Run through the beginning of the input file to check for leading "newdmformat"
 * characters if they are present then indicate this in the return value.
 */

int
isNonInteractDMFormatProbablyNewDMFormat( IFDp, lineBufferStructp )
struct  inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
{
int	numLinesRead = 0, DMFORMATfileTypeRejected = FALSE ;

	/* Read the top of the file line by line */

	while( ++numLinesRead < 50 &&
	       DMFORMATfileTypeRejected == FALSE &&
	       getLine( IFDp, lineBufferStructp, ANYLINETYPE, UNDEFFORMATINPUT ) == OK )
	{
	    if( isCharInSet( *(lineBufferStructp -> lineBufferp),
	    		     ALLBULKLOADCHARS ) == TRUE )
	    {
		DMFORMATfileTypeRejected = TRUE ;
	    }
	    
	    *(lineBufferStructp -> lineBufferp) = '\0' ;
	}

	/* Return the input file pointer to the first line.. */
	rewind( IFDp -> inFilep ) ;

	return( DMFORMATfileTypeRejected ) ;
}

/* 
 * Run through the beginning of the input file to check for leading "newdmformat"
 * characters if they are present then the user is given the opportunity
 * to correct the selection of the dmformat file format.
 */

void
checkInteractInFileTypeIsNotNewDMFormat( IFDp, lineBufferStructp, inFileTypep )
struct  inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
int  *	inFileTypep ;
{
int	numLinesRead = 0, DMFORMATfileTypeRejected = FALSE ;

	/* Read the top of the file line by line */

	while( ++numLinesRead < 50 &&
	       DMFORMATfileTypeRejected == FALSE &&
	       getLine( IFDp, lineBufferStructp, ANYLINETYPE, UNDEFFORMATINPUT ) == OK )
	{
	    if( isCharInSet( *(lineBufferStructp -> lineBufferp),
	    		     ALLBULKLOADCHARS ) == TRUE )
	    {
		DMFORMATfileTypeRejected = TRUE ;

	    	if( enterYesNo( lineBufferStructp -> lineBufferp,
"Although the \"newdmformat\" statement is missing from the top\n\
of the input file, this line appears to use this format\n\
File is in fact in \"newdmformat\" (y/n) ? " ) == 'y' )
		    *inFileTypep = NEWDMFORMATINPUT ;
		else
		    DMFORMATfileTypeRejected = FALSE ;
	    }
	    
	    *(lineBufferStructp -> lineBufferp) = '\0' ;
	}

	/* Return the input file pointer to the first line.. */
	rewind( IFDp -> inFilep ) ;

	return ;
}


/*
 * Check for the input file type indication that is present in
 * output files from "newdmdiff"..
 */

int
retrieveFileType( IFDp, lineBufferStructp, inFileTypep )
struct  inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
int  *	inFileTypep ;
{
int	ret_val = OK ;

	if( getLine( IFDp, lineBufferStructp, ANYLINETYPE, UNDEFFORMATINPUT ) != OK )
	{
	    fprintf( stderr, "Input file is empty\n" );
	    ret_val = NOTOK ;
	}
	else
	{
	    if( determineLineType( lineBufferStructp->lineBufferp ) == NEWDMFORMATLINETYPE )
	    {
		*inFileTypep = NEWDMFORMATINPUT ;

		/* Line buffer has been used.. */
	    	*(lineBufferStructp -> lineBufferp) = '\0' ;
	    }
	    else
	    	*inFileTypep = DMFORMATFILEINPUT ;
	}

	return( ret_val ) ;
}

/*
 * Extract the characters that follow the "fromChar" (excluding initial
 * whitespace characters.)
 */

int
extractFromChar( IFDp, lineBufferp, fromChar, tailStrpp )
struct	inFileDetails * IFDp ;
char *	lineBufferp ;
char	fromChar  ;
char ** tailStrpp ;
{
int	ret_val = OK ;
char *  stringp ;

	if( (stringp = index( lineBufferp, fromChar )) == NULL )
	{
#ifdef DEBUG
fprintf( stderr, "Input file problem '%c' missing within\n%s\n", fromChar, lineBufferp );
#endif
	    reportParseError( IFDp, "Input file problem '=' missing within", lineBufferp );

	    ret_val = NOTOK ;
	}
	else
    	{
	    stringp++ ; /* Passed the "fromChar" */
		
	    /* Increment pointer passed any white space characters.. */
	    while( isspace( *stringp ) )
	    	stringp++ ;
		
	    if( ((*tailStrpp) = strdup( stringp )) == NULL )
	    {
#ifdef DEBUG
fprintf( stderr, "Memory allocation problem in `extractFromChar`\n" );
#endif
	    	reportFatalParseError( IFDp , "Memory allocation problem in `extractFromChar`", NULLCP );
		ret_val = NOTOK ;
	    }
	}

	return( ret_val );
}


void
removeTrailWspce( stringp )
char *	stringp ;
{
char *	strcp ;

	strcp = stringp + strlen( stringp ) - 1 ;

	while( isspace( *strcp ) )
	{
	    *strcp-- = '\0' ;
	}
}


/*
 * Convert a DN string to the standardised form in "@" notation.
 * The parameter may be in UFN or "@" notation.
 * The UFN format assumes that the attribute types have been included..
 * e.g. "cn=Paul Barker, ou= Computer Science, o=UCL, c=GB"
 */

int
standardiseDNstring( IFDp, DNstrpp )
struct	inFileDetails     * IFDp ;
char **	DNstrpp ; /* standardised string returned to calling procedure */
{
int	ret_val = OK, lastUFNCompFound = FALSE ;
char *	DNstrcp = NULL  ;
char *	delimcp = NULL  ;
int	currReturnStrLen, newTotalStrLen ;


	/* Remove trailing whitespace characters */
	removeTrailWspce( *DNstrpp ) ;

	/* Determine whether DN string is in the "@" notation already */
	if( index( *DNstrpp, RDNDELIMITER )    == NULL &&
	    index( *DNstrpp, UFNRDNDELIMITER ) != NULL   )
	{
	    /* UFN format string */
	
	    /* Copy the source DN string */
	    if( (DNstrcp = strdup( *DNstrpp )) == NULL )
	    {
#ifdef DEBUG
fprintf( stderr, "Memory allocation failure in `standardiseDNstring'\n" );
#endif
		reportFatalParseError( IFDp, "Memory allocation failure in `standardiseDNstring'", NULLCP ) ;
		ret_val = NOTOK ;
	    }
	    else
	    {
		currReturnStrLen = strlen( *DNstrpp ) ;

		**DNstrpp = '\0' ; /* Clear the source DN string */
	    
	    	/* .. assume UFN format */
	    	delimcp = rindex( DNstrcp, UFNRDNDELIMITER ) ;
		
		while( delimcp != NULL )
		{
		    if( lastUFNCompFound == FALSE ) 
		    	*delimcp++ = '\0' ; /* Extract off the right most UFN component */

		    /* Ensure that the return string is large enough.. */

		    newTotalStrLen = strlen( *DNstrpp ) + strlen( delimcp ) + 2 ;
		    
		    if( currReturnStrLen < newTotalStrLen )
		    {
		    	/* ..no so realloc().. */
			if((*DNstrpp = realloc( (char *)(*DNstrpp),
								(unsigned)newTotalStrLen )) == NULL )
			{
#ifdef DEBUG
fprintf( stderr, "Realloc failure in `standardiseDNstring'\n" );
#endif
			    reportFatalParseError( IFDp, "Realloc failure in `standardiseDNstring'", NULLCP ) ;

			    ret_val = NOTOK ;
			    break ;
			}
#ifdef DEBUG
else
fprintf( stderr, "Reallocing %d\n", newTotalStrLen) ;
#endif
			currReturnStrLen = newTotalStrLen ;
		    }
		
		    /* The "@" is not prepended for the first component.. */
		    if( **DNstrpp != '\0' )
		    	strcat( *DNstrpp, "@" ) ;
			
		    strcat( *DNstrpp, delimcp ) ;


		    /* Find the next rightmost UFN RDN delimiter.. */
		    if( (delimcp = rindex( DNstrcp, UFNRDNDELIMITER ) ) == NULL )
		    {
		    	/* Check to see if the last UFN component has been extracted.. */
		    	if( lastUFNCompFound == FALSE )
			{
			   delimcp = DNstrcp ;  /* final component */
			   lastUFNCompFound = TRUE ;
			}
			else
			   delimcp = NULL ;
		    }

		} /* while */

	    	free( DNstrcp ) ;
	    }
	}

	return( ret_val );
}

/*
 * Transfer the remaining lines from the current input/parse file to the output parse file.
 */
void
transferInFileToOutParseFile( IFDp, lineBufferStructp )
struct	inFileDetails 	 * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
{
	/* Transfer the parse file.. */

	/* Clear the current contents */
	fclose( IFDp -> parseOutputFilep ) ; 
	IFDp -> parseOutputFilep = fopen( IFDp -> parseOutputFilename, "w" ) ;
	
	if( *( lineBufferStructp -> lineBufferp ) != '\0' )
	    fprintf( IFDp -> parseOutputFilep, "%s", lineBufferStructp -> lineBufferp ) ;
	
	while( fgets( lineBufferStructp -> lineBufferp, MAXLINELEN, IFDp -> parseInputFilep ) != NULL )
	{
	    fprintf( IFDp -> parseOutputFilep, "%s", lineBufferStructp -> lineBufferp ) ;

	    *( lineBufferStructp -> lineBufferp ) = '\0' ; /* Line has been used */
	}

	return ;
}


void
fatalTransferRestOfInputToProblems( IFDp, OFD, lineBufferStructp, errorMessagecp )
struct	inFileDetails * IFDp ;
struct	outputFileDetails  OFD[] ;
struct	lineBufferStruct * lineBufferStructp ;
char *	errorMessagecp ;
{
	if( errorMessagecp != NULLCP )
	    fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "# %s\n", errorMessagecp ) ;

	if( OFD[ OUTPUTPROBLEMFILE ].RootedAtOutput == FALSE &&
	    OFD[ OUTPUTPROBLEMFILE ].RootedAtDNcp != NULLCP )
	{
	    fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "\n%s\n", OFD[ OUTPUTPROBLEMFILE ].RootedAtDNcp ) ;

	    if( OFD[ OUTPUTPROBLEMFILE ].SubDirListResolvedParentDNcp != NULLCP )
		fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "# DN : %s\n", OFD[ OUTPUTPROBLEMFILE ].SubDirListResolvedParentDNcp ) ;
	    
	    OFD[ OUTPUTPROBLEMFILE ].RootedAtOutput = TRUE ;
	}

	if( *( lineBufferStructp -> lineBufferp ) != '\0' )
	{
	    fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "%s", lineBufferStructp -> lineBufferp ) ;
	
	    *( lineBufferStructp -> lineBufferp ) = '\0' ; /* Line has been transferred to parse file */
	}

	setUpParseFileDetails( IFDp, lineBufferStructp, PARSEINPUTFILEINPARSE1OUT ) ; /* Read from the input file */

	/* Transfer the remaining lines of the input file.. */
	
	while( fgets( lineBufferStructp -> lineBufferp, MAXLINELEN, IFDp -> parseInputFilep ) != NULL )
	{
	    fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "%s", lineBufferStructp -> lineBufferp ) ;

	    *( lineBufferStructp -> lineBufferp ) = '\0' ; /* Line has been used */
	}

	return ;
}


/*
 * If in non-interactive mode a parse failure or a fatal failure when in interactive
 * mode (e.g. malloc failure) causes the lines currently buffered
 * in the parse file plus all remainging un-read input lines to be transferred
 * to the problems file.
 */

void
transferParseFileAndRestOfInputToProblems( operationalMode, IFDp, OFD, lineBufferStructp )
int	operationalMode ;
struct	inFileDetails * IFDp ;
struct	outputFileDetails  OFD[] ;
struct	lineBufferStruct * lineBufferStructp ;
{
	if( operationalMode == NON_INTERACT_OPER_MODE )
	    reportFatalParseError( IFDp, "Parse recovery requires user intervention but in non-interactive mode", NULLCP );

	if( OFD[ OUTPUTPROBLEMFILE ].RootedAtOutput == FALSE &&
	    OFD[ OUTPUTPROBLEMFILE ].RootedAtDNcp != NULLCP )
	{
	    fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "%s\n", OFD[ OUTPUTPROBLEMFILE ].RootedAtDNcp ) ;
	    if( OFD[ OUTPUTPROBLEMFILE ].SubDirListResolvedParentDNcp != NULLCP )
		fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "# DN : %s\n", OFD[ OUTPUTPROBLEMFILE ].SubDirListResolvedParentDNcp ) ;

	    OFD[ OUTPUTPROBLEMFILE ].RootedAtOutput = TRUE ;
	}
	
	if( *( lineBufferStructp -> lineBufferp ) != '\0' )
	{
	    fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "%s", lineBufferStructp -> lineBufferp ) ;
	
	    *( lineBufferStructp -> lineBufferp ) = '\0' ; /* Line has been transferred to parse file */
	}

	setUpParseFileDetails( IFDp, lineBufferStructp, SWAPPARSEFILES ) ; /* Read from the last updated parse file */

	/* Transfer the parse file.. */
	
	while( fgets( lineBufferStructp -> lineBufferp, MAXLINELEN, IFDp -> parseInputFilep ) != NULL )
	{
	    fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "%s", lineBufferStructp -> lineBufferp ) ;

	    *( lineBufferStructp -> lineBufferp ) = '\0' ; /* Line has been used */
	}

	setUpParseFileDetails( IFDp, lineBufferStructp, PARSEINPUTFILEINPARSE1OUT ) ; /* Read from the input file */

	/* Transfer the remaining lines of the input file.. */
	
	while( fgets( lineBufferStructp -> lineBufferp, MAXLINELEN, IFDp -> parseInputFilep ) != NULL )
	{
	    fprintf( OFD[ OUTPUTPROBLEMFILE ].outFilep, "%s", lineBufferStructp -> lineBufferp ) ;

	    *( lineBufferStructp -> lineBufferp ) = '\0' ; /* Line has been used */
	}

	return ;
}


/*
 * When retrieving the source name a check is required to 
 * determine whether the source name contains the DN for
 * a specific dataSource entry.
 */

int
retrieveSource( operationalMode, IFDp, OFD, lineBufferStructp, 
			sourcecpp, dataSourceEntryDNp, DNIncludedSourceNamecpp )
int	operationalMode ;
struct	inFileDetails * IFDp ;
struct	outputFileDetails  OFD[] ;
struct	lineBufferStruct * lineBufferStructp ;
char **	sourcecpp ;
DN   *	dataSourceEntryDNp ;
char **	DNIncludedSourceNamecpp ;
{
int	ret_val = OK, res, loopExit = FALSE ;
char *	sourceBuffercp, *cp ;
RDN	lastSourceRDN = NULLRDN ;
static	AttributeType dataSourceat = NULLAttrT ;

	while( loopExit == FALSE )
	{
	    if( (ret_val = getLine( IFDp, lineBufferStructp, ANYLINETYPE, UNDEFFORMATINPUT )) != OK )
	    {
		loopExit = TRUE ; /* No data in file !! */
	    }
	    else
	    {
	    	if( determineLineType( lineBufferStructp -> lineBufferp ) != SOURCELINETYPE )
	    	{
	    	    reportParseError( IFDp, "Input file does not contain a source line", NULLCP );
	    	}
	    	else
	    	{
	    	    /* Extract the source string from the source line.. */
	    	    if( (res = extractFromChar( IFDp, lineBufferStructp->lineBufferp, '=', &sourceBuffercp )) == OK )
		    {
			/* Check to see if this is a special case with the
			   source name containing a dataSource entry DN..  */

			*dataSourceEntryDNp      = NULLDN ;
			*DNIncludedSourceNamecpp = NULLCP ;

			if( sourceBuffercp != NULLCP && (cp = strrchr( sourceBuffercp, '=' )) != NULLCP ) {

			    if( (*dataSourceEntryDNp = str2dn( sourceBuffercp )) != NULLDN ) {

				if( dataSourceat == NULLAttrT )
				    dataSourceat =  str2AttrT( BL_DATA_SOURCE ) ;

				if( (lastSourceRDN = findLastRDN( *dataSourceEntryDNp )) != NULLRDN &&
				    lastSourceRDN -> rdn_at != NULLAttrT &&
				    AttrT_cmp( lastSourceRDN -> rdn_at, dataSourceat ) == 0 ) {

				    /* The source name specifies the DN of a full dataSource entry */

				    /* Take the dataSource value as the source name.. */
				    cp++ ; /* skip the equals */

				    while( isspace( *cp ) )
					cp++ ; /* Just in case there are any leading spaces */

				    *sourcecpp = strdup( cp ) ;

				    *DNIncludedSourceNamecpp = sourceBuffercp ;
				}
				else {
				    dn_free( *dataSourceEntryDNp ) ;
				    *dataSourceEntryDNp = NULLDN ;
				}
			    }
			}

			if( *dataSourceEntryDNp == NULLDN )
			    *sourcecpp = sourceBuffercp ;

			loopExit = TRUE ;
		    }
		    
		    /* Line buffer has been used or transferred to a parse file.. */
	    	    *(lineBufferStructp -> lineBufferp) = '\0' ;
		}
	        
		if( loopExit == FALSE )
		{
		    if( operationalMode == INTERACT_OPER_MODE )
		    {
		    	closeParseFiles( IFDp ) ;
		
		    	editFile( IFDp -> parseOutputFilename ) ;

		    	setUpParseFileDetails( IFDp, lineBufferStructp, SWAPPARSEFILES ) ;
		    }
		    else
		    {
		    	/* Non-interactive mode (FATAL)*/
			transferParseFileAndRestOfInputToProblems( operationalMode, IFDp, OFD, lineBufferStructp ) ;
			loopExit = TRUE ; ret_val = NOTOK ;
		    }
		}
	    }
	}

	/* The user's file is again specified for input.. */
	setUpParseFileDetails( IFDp, lineBufferStructp, PARSEINPUTFILEINPARSE1OUT ) ;

	return( ret_val ) ;
}


int
retrieveDN( IFDp, lineBufferStructp, DNstrpp, lineType )
struct	inFileDetails    * IFDp ;
struct	lineBufferStruct * lineBufferStructp ;
char **	DNstrpp  ;
int	lineType ;
{
int	ret_val = OK ;

	if( getLine( IFDp, lineBufferStructp, lineType, UNDEFFORMATINPUT ) != OK )
	{
	    ret_val = NOTOK ;
	}
	else
	{
	    /* Extract the DN string from the source line.. */
	    ret_val = extractFromChar( IFDp, lineBufferStructp->lineBufferp, '=', DNstrpp ) ;

	    if( ret_val == OK )
	    {
		/* Standardise the string DNs format to the "@" notation */
		ret_val = standardiseDNstring( IFDp, DNstrpp ) ;
	    }

	    /* Line buffer has been used.. */
	    *(lineBufferStructp -> lineBufferp) = '\0' ;
	}

	return( ret_val ) ;
}

/*
 * Note: on a fatal interrupt the procedure must ensure that any remaining
 * unprocessed lines are passed to the PROBLEMs file before returning. 
 * The procedures state can only be used if the interrupt occurred when in 
 * some sub-procedure (hence the usage of 'fatalInterruptCanBeHandled').
 */

void
processInputFileRootedAts( operationalMode, inFileType, IFDp, OFD, lineBufferStructp, 
		sourceNamecp, dataSourceEntryDN, managementOperation, fatalInterrupt )
int	operationalMode,    inFileType ;
struct	inFileDetails     * IFDp  ;
struct	outputFileDetails   OFD[] ;
struct	lineBufferStruct  * lineBufferStructp ;
char * 	sourceNamecp ;
int	managementOperation ;
int	fatalInterrupt ;
{
char *  rootedAtDNcp = NULLCP ;
static	struct 	rootedAtAttributes * baseRAAp ;
int	res, loopExit = FALSE, parseLoopExit, atleastOneRootedAtFound=FALSE ;

static	int 			   fatalInterruptCanBeHandled = FALSE ; /* Indicates whether the following are valid.. */
static	struct inFileDetails * 	   buffFatalIFDp ;
static	struct outputFileDetails * buffFatalOFDp ;
static	struct lineBufferStruct  * buffFatalLineBufferStructp ;

	if( fatalInterrupt == TRUE )
	{
	    if( fatalInterruptCanBeHandled == TRUE )
	    {
		/* Transfer any lines that are part of the RDN block that was being processed.. */
		if( baseRAAp != NULLRAA )
		    transferRootedAtRDNattrs( buffFatalOFDp, OUTPUTPROBLEMFILE, baseRAAp, "*** FATAL ERROR DSA FAILURE *** FATAL ERROR DSA FAILURE ***",
	  		    	"*** FATAL ERROR DSA FAILURE *** FATAL ERROR DSA FAILURE ***" ) ;

		/* Transfer any remaining un-read lines.. */
		fatalTransferRestOfInputToProblems( buffFatalIFDp, buffFatalOFDp, buffFatalLineBufferStructp, NULLCP ) ;

		fatalInterruptCanBeHandled = FALSE ;
		
	    	/* Remove temporary files before leaving.. */
	    	if( buffFatalIFDp -> parse1Filename[ 0 ] != '\0'  && strncmp( buffFatalIFDp -> parse1Filename, PARSEAFILENAME, 12 ) == 0 )
		    unlink( buffFatalIFDp -> parse1Filename ) ;

	    	if( buffFatalIFDp -> parse2Filename[ 0 ] != '\0'  && strncmp( buffFatalIFDp -> parse2Filename, PARSEAFILENAME, 12 ) == 0 )
		    unlink( buffFatalIFDp -> parse2Filename ) ;
		     
		if( strncmp( buffFatalOFDp[ OUTPUTDRAFTFILE ].outputFilename, DRAFTFILENAME, 13 ) == 0 )
		    unlink( buffFatalOFDp[ OUTPUTDRAFTFILE ].outputFilename ) ;
	    }

	    return ;
	}

	buffFatalIFDp = IFDp ;
	buffFatalOFDp = OFD ;
	buffFatalLineBufferStructp = lineBufferStructp ;

	fatalInterruptCanBeHandled = TRUE ;

	baseRAAp = NULLRAA ;

	/* rootedAt loop */
	while( loopExit == FALSE )
	{
	    parseLoopExit = FALSE ;

	    setUpOutFilesRA(  OFD, NULLCP ) ;
	    setUpOutFilesRDN( OFD, NULLCP ) ;
	
	    while( parseLoopExit == FALSE )
	    {	    
	    	/* Prepare for the next "rootedAt" loop iteration (if any).. */
		res = retrieveDN( IFDp, lineBufferStructp, &rootedAtDNcp, ROOTEDATLINETYPE ) ;

 		/* Note: failure to find a "rootedAt" statement is only
 			 a problem if we have not found any previously */
 
 		if( res == NOTOK && IFDp -> fatalErrorOccurred == FALSE &&
 			atleastOneRootedAtFound == FALSE )
 		    reportFatalParseError( IFDp, "Failed to find/parse the rootedAt statement", NULLCP );
 
 	    	if( (res == NOTOK && IFDp->errorOccurred == FALSE ) || IFDp -> fatalErrorOccurred == TRUE )
		{
		    if( IFDp -> fatalErrorOccurred == TRUE )
		    	transferParseFileAndRestOfInputToProblems( operationalMode, IFDp, OFD, lineBufferStructp ) ;

		    parseLoopExit = TRUE ; loopExit = TRUE ;
		}
		else if( IFDp -> errorOccurred == TRUE )
		{
		    if( operationalMode == INTERACT_OPER_MODE )
		    {
		    	closeParseFiles( IFDp ) ;
		
		    	editFile( IFDp -> parseOutputFilename ) ;

		    	setUpParseFileDetails( IFDp, lineBufferStructp, SWAPPARSEFILES ) ;
		    }
		    else
		    {
		    	/* Non-interactive mode (FATAL)*/
			transferParseFileAndRestOfInputToProblems( operationalMode, IFDp, OFD, lineBufferStructp ) ;
		    	parseLoopExit = TRUE ; loopExit = TRUE ;
		    }
		}
		else {
 		    atleastOneRootedAtFound = TRUE ;
		    parseLoopExit = TRUE ;
		}
	    }

	    if( res == OK && loopExit == FALSE && IFDp -> fatalErrorOccurred == FALSE )
	    {
		/* The user's file is again specified for input.. */
		setUpParseFileDetails( IFDp, lineBufferStructp, PARSEINPUTFILEINPARSE1OUT ) ;

		setUpOutFilesRA( OFD, rootedAtDNcp ) ;
	    }

	    /* RDN block loop */
	    while( res == OK && loopExit == FALSE && IFDp -> fatalErrorOccurred == FALSE )
	    {
		parseLoopExit = FALSE ;
	    	
		while( parseLoopExit == FALSE )
	    	{
		    /* Buffer a single RDN block.. */
		    res = bufferRDNplusAttrs( IFDp, lineBufferStructp, &baseRAAp, inFileType, rootedAtDNcp ) ;

		    /* .. returns OK, NOTOKALLDATAREAD when RDN block fully read
		       and NOTOK on some failure e.g. of a malloc */

		    if( IFDp -> fatalErrorOccurred == TRUE )
		    {
			transferParseFileAndRestOfInputToProblems( operationalMode, IFDp, OFD, lineBufferStructp ) ;
		    	parseLoopExit = TRUE ; loopExit = TRUE ;
		    }
		    else if( res == NOTOK || IFDp -> errorOccurred == TRUE )
		    {
			if( operationalMode == INTERACT_OPER_MODE )
			{
			    freeRootedAtAttrs_list( &baseRAAp ) ;
		    
		    	    closeParseFiles( IFDp ) ;
		
		    	    editFile( IFDp -> parseOutputFilename ) ;

			    setUpParseFileDetails( IFDp, lineBufferStructp, SWAPPARSEFILES ) ;
			}
			else
			{
			    transferParseFileAndRestOfInputToProblems( operationalMode, IFDp, OFD, lineBufferStructp ) ;
			    parseLoopExit = TRUE ; loopExit = TRUE ;
			}
		    }
		    else parseLoopExit = TRUE ;
		}
		
		if( res == OK && loopExit == FALSE && baseRAAp != NULLRAA )
		{
		    /* The user's file is again specified for input.. */
		    setUpParseFileDetails( IFDp, lineBufferStructp, PARSEINPUTFILEINPARSE1OUT ) ;

		    if( managementOperation == TRUE )
		    {
		    	res = attemptMatchAgainstDirectory( &baseRAAp, OFD, sourceNamecp ) ;
		    }
		    else
		    {
		    	res = updateDirectory( &baseRAAp, inFileType, operationalMode, OFD, sourceNamecp, dataSourceEntryDN ) ;
		    }

		    if( baseRAAp != NULLRAA )
			freeRootedAtAttrs_list( &baseRAAp ) ;

		    flushOutFiles( OFD ) ;
		}

	    } /* while loop */

	    if( rootedAtDNcp != NULLCP )
	    {
	    	free( rootedAtDNcp ) ;
		rootedAtDNcp = NULLCP ;
	    }
	}

	fatalInterruptCanBeHandled = FALSE ;
	
	return ;
}


void
processInputFile( operationalMode, file_bufferp, managementOperation )
int	operationalMode ;
char *	file_bufferp ;
int	managementOperation ; /* TRUE if called from bulk management, else called from bulkload */
{
struct	lineBufferStruct  loadLineBufferStruct ;
struct	inFileDetails  IFD, * IFDp = &IFD ;
struct	outputFileDetails outFileDetails[MAXOUTPUTFILEINDEX+1] ;
char *	sourceNamecp ;
int	inFileType, lenSourceNamecp ;
DN	dataSourceEntryDN      = NULLDN ; /* For the special case when the     */
char *	DNIncludedSourceNamecp = NULLCP ; /* source name holds a dataSource DN */

	if( initialiseInFiles( IFDp, file_bufferp, PARSEAFILENAME, PARSEBFILENAME ) == OK &&
	    initialiseOutFiles( file_bufferp, outFileDetails )                      == OK   )
	{
	    if( (loadLineBufferStruct.lineBufferp = malloc( INITLINEBUFLEN )) != NULL )
	    {
		*(loadLineBufferStruct.lineBufferp) = '\0' ;
		loadLineBufferStruct.bufferSize = INITLINEBUFLEN ;

		if( retrieveFileType( IFDp, &loadLineBufferStruct, &inFileType ) == OK )
		{
		    if( operationalMode == NON_INTERACT_OPER_MODE && inFileType == DMFORMATFILEINPUT &&
		        isNonInteractDMFormatProbablyNewDMFormat( IFDp, &loadLineBufferStruct ) == TRUE )
		    {
		    	/* We require interactive resolution of the probable missing "newdmformat" header */

			fatalTransferRestOfInputToProblems( IFDp, outFileDetails, &loadLineBufferStruct,
 				"File believed to be in \"newdmformat\" requires interactive intervention" ) ;
		    }
		    else
		    {
		    	if( operationalMode == INTERACT_OPER_MODE && inFileType == DMFORMATFILEINPUT )
			    checkInteractInFileTypeIsNotNewDMFormat( IFDp, &loadLineBufferStruct, &inFileType ) ;
		
		    	setUpParseFileDetails( IFDp, &loadLineBufferStruct, PARSEINPUTFILEINPARSE1OUT ) ;

		    	if( retrieveSource( operationalMode, IFDp, outFileDetails, &loadLineBufferStruct, &sourceNamecp, &dataSourceEntryDN, &DNIncludedSourceNamecp ) == OK )
		    	{
			    /* Check for trailing '\n' in the source name.. */
			    lenSourceNamecp = strlen( sourceNamecp ) ;
			    if( *(sourceNamecp + lenSourceNamecp - 1 ) == '\n' )
			    	*(sourceNamecp + lenSourceNamecp - 1 ) =  '\0' ;

			    prepareOutFiles( outFileDetails, inFileType, 
				(DNIncludedSourceNamecp == NULLCP) ? sourceNamecp : DNIncludedSourceNamecp ) ;

			    processInputFileRootedAts( operationalMode, inFileType, IFDp, outFileDetails,
					&loadLineBufferStruct, sourceNamecp, dataSourceEntryDN, managementOperation, FALSE ) ;

			    /* Free source name related storage.. */
			    if( DNIncludedSourceNamecp ) (void)free( DNIncludedSourceNamecp ) ;

			    if( dataSourceEntryDN != NULLDN ) dn_free( dataSourceEntryDN ) ;

			    free( sourceNamecp ) ;
		    	}
		    }

		    free( loadLineBufferStruct.lineBufferp ) ;
		}
	    }
	    else fprintf( stderr, "Memory allocation failure when creating the line buffer\n" );

	    /* Remove temporary files before leaving.. */
	    if( IFDp -> parse1Filename[ 0 ] != '\0' && strncmp( IFDp -> parse1Filename, PARSEAFILENAME, 12 ) == 0 )
		unlink( IFDp -> parse1Filename ) ;

	    if( IFDp -> parse2Filename[ 0 ] != '\0'  && strncmp( IFDp -> parse2Filename, PARSEAFILENAME, 12 ) == 0 )
		unlink( IFDp -> parse2Filename ) ;

	    if( strncmp( outFileDetails[ OUTPUTDRAFTFILE ].outputFilename, DRAFTFILENAME, 13 ) == 0 )
		unlink( outFileDetails[ OUTPUTDRAFTFILE ].outputFilename ) ;
	}

	
	return ;
}


int
processRAandRDNEditFile( inFileType, OFD, preEditRAApp )
int	inFileType ;
struct	outputFileDetails   OFD[] ;
struct	rootedAtAttributes ** preEditRAApp ;
{
int	ret_val = NOTOK, parseLoopExit = FALSE ;
struct	inFileDetails IFD, *IFDp = &IFD ;
char *  rootedAtDNcp = NULL;	
struct 	rootedAtAttributes *  baseRAAp  = NULL ;
int	res ;
struct	lineBufferStruct  draftLineBufferStruct ;

	if( (draftLineBufferStruct.lineBufferp = malloc( INITLINEBUFLEN )) != NULL &&
	    initialiseInFiles( IFDp, OFD[ OUTPUTDRAFTFILE ].outputFilename, PARSECFILENAME, PARSEDFILENAME ) == OK )
	{
	    *(draftLineBufferStruct.lineBufferp) = '\0' ;
	    draftLineBufferStruct.bufferSize = INITLINEBUFLEN ;

	    parseLoopExit = FALSE ;

	    while( parseLoopExit == FALSE )
	    {
	    	res = retrieveDN( IFDp, &draftLineBufferStruct, &rootedAtDNcp, ROOTEDATLINETYPE ) ;

#ifdef DEBUG
fprintf( stderr, "Retrieved DN %s\n", rootedAtDNcp ) ;
#endif
	    	if( res == NOTOK || IFDp -> fatalErrorOccurred == TRUE )
		    parseLoopExit = TRUE ;

		else if( IFDp -> errorOccurred == TRUE )
		{
		    closeParseFiles( IFDp ) ;
		
		    editFile( IFDp -> parseOutputFilename ) ;

		    setUpParseFileDetails( IFDp, &draftLineBufferStruct, SWAPPARSEFILES ) ;
		}
		else parseLoopExit = TRUE ;
	    }


	    if( res == OK && IFDp -> fatalErrorOccurred == FALSE )
	    {
	    	parseLoopExit = FALSE ;

		/* Transfer the RDN and attribute lines to the output file */
		transferInFileToOutParseFile( IFDp, &draftLineBufferStruct ) ;

		setUpParseFileDetails( IFDp, &draftLineBufferStruct, SWAPPARSEFILES ) ;
		
		while( parseLoopExit == FALSE )
	    	{	    
		    setUpOutFilesRA( OFD, rootedAtDNcp ) ;
		
	            res = bufferRDNplusAttrs( IFDp, &draftLineBufferStruct, &baseRAAp, inFileType, rootedAtDNcp ) ;

	            /* .. returns OK, NOTOKALLDATAREAD when RDN block fully read
		       and NOTOK on some failure e.g. of a malloc */

		    if( IFDp -> fatalErrorOccurred == TRUE )
		    {
		    	parseLoopExit = TRUE ;
		    }
		    else if( res == NOTOK || IFDp -> errorOccurred == TRUE )
		    {
			freeRootedAtAttrs_list( &baseRAAp ) ;

		    	closeParseFiles( IFDp ) ;
		
		    	editFile( IFDp -> parseOutputFilename ) ;

		    	setUpParseFileDetails( IFDp, &draftLineBufferStruct, SWAPPARSEFILES ) ;
		    }
		    else
		    {
		    	parseLoopExit = TRUE ;

		    	setUpOutFilesRDN( OFD, baseRAAp -> rdnLinecp ) ;

	    	    	*preEditRAApp = baseRAAp ;
		    	ret_val = OK ;
	    	    }
	    	}

	    	if( rootedAtDNcp != NULL )
	    	    free( rootedAtDNcp ) ;

	    	finishIFD( IFDp ) ;

	    	free( draftLineBufferStruct.lineBufferp ) ;
	    }
	}
	
	return( ret_val ) ;
}


int
processAttributeEditFile( inFileType, OFD, entryRDN, preEditAttrLinepp, singleAttributeLineOnly )
int	inFileType ;
struct	outputFileDetails   OFD[] ;
RDN	entryRDN ;
struct	attributeLine 	 ** preEditAttrLinepp ;
int	singleAttributeLineOnly ; /* FALSE if list of attributes required */
{
int	ret_val = NOTOK, parseLoopExit = FALSE ;
struct	inFileDetails IFD, *IFDp = &IFD ;
struct 	attributeLine *  baseAttrLinep  = NULL ;
int	res ;
struct	lineBufferStruct  draftLineBufferStruct ;

	if( (draftLineBufferStruct.lineBufferp = malloc( INITLINEBUFLEN )) != NULL &&
	    initialiseInFiles( IFDp, OFD[ OUTPUTDRAFTFILE ].outputFilename, PARSECFILENAME, PARSEDFILENAME ) == OK )
	{
	    *(draftLineBufferStruct.lineBufferp) = '\0' ;
	    draftLineBufferStruct.bufferSize = INITLINEBUFLEN ;

	    parseLoopExit = FALSE ;

	    while( parseLoopExit == FALSE )
	    {
	    	res = bufferAttrs( IFDp, &draftLineBufferStruct, entryRDN,
	    			&baseAttrLinep, inFileType, ATTSLINETYPE ) ;

	    	/* .. returns OK, NOTOKALLDATAREAD when RDN block fully read
	    	   and NOTOK on some fatal failure e.g. of a malloc */

	    	if( res == NOTOK || IFDp -> fatalErrorOccurred == TRUE )
		    parseLoopExit = TRUE ;

		else if( IFDp -> errorOccurred == TRUE )
		{
		    free_attrLine_list( baseAttrLinep ) ;
		    
		    closeParseFiles( IFDp ) ;
		
		    editFile( IFDp -> parseOutputFilename ) ;

		    setUpParseFileDetails( IFDp, &draftLineBufferStruct, SWAPPARSEFILES ) ;
		}
		else 
		{
		    parseLoopExit = TRUE ;

	    	    *preEditAttrLinepp = baseAttrLinep ;

		    ret_val = OK ;
	    	}
	    }

	    finishIFD( IFDp ) ;

	    free( draftLineBufferStruct.lineBufferp ) ;
	}

	return( ret_val ) ;
}


